Package: llm.api
Title: Minimal LLM Chat Interface
Version: 0.1.4
Date: 2026-05-24
Authors@R: c(
    person("Troy", "Hernandez", role = c("aut", "cre"),
           email = "troy@cornball.ai",
           comment = c(ORCID = "0009-0005-4248-604X")),
    person("cornball.ai", role = "cph"),
    person("ellmer team", role = "cph",
           comment = "API design derived from ellmer"))
Description: A minimal-dependency client for Large Language Model chat APIs.
    Supports 'OpenAI' <https://openai.com/>, 'Anthropic' 'Claude'
    <https://claude.com/>, 'Moonshot' 'Kimi' <https://www.moonshot.ai/>,
    'Ollama' <https://ollama.com/>, and other 'OpenAI'-compatible
    endpoints. Includes an agent loop with tool use and a 'Model Context
    Protocol' client <https://modelcontextprotocol.io/>. API design is
    derived from the 'ellmer' package, reimplemented with only base R,
    'curl', and 'jsonlite'.
License: MIT + file LICENSE
URL: https://github.com/cornball-ai/llm.api
BugReports: https://github.com/cornball-ai/llm.api/issues
Encoding: UTF-8
Imports: curl, jsonlite
Suggests: tinytest
NeedsCompilation: no
Packaged: 2026-05-24 22:53:00 UTC; troy
Author: Troy Hernandez [aut, cre] (ORCID:
    <https://orcid.org/0009-0005-4248-604X>),
  cornball.ai [cph],
  ellmer team [cph] (API design derived from ellmer)
Maintainer: Troy Hernandez <troy@cornball.ai>
Depends: R (>= 3.5.0)
Repository: CRAN
Date/Publication: 2026-05-25 15:50:14 UTC
Built: R 4.6.0; ; 2026-05-25 18:13:30 UTC; unix

Package: llm.api
Title: Minimal LLM Chat Interface
Version: 0.1.9
Date: 2026-08-04
Authors@R: c(
    person("Troy", "Hernandez", role = c("aut", "cre"),
           email = "troy@cornball.ai",
           comment = c(ORCID = "0009-0005-4248-604X")),
    person("cornball.ai", role = "cph"),
    person("ellmer team", role = "cph",
           comment = "API design derived from ellmer"))
Description: A minimal-dependency client for Large Language Model chat APIs.
    Supports 'OpenAI' <https://openai.com/>, 'Anthropic' 'Claude'
    <https://claude.com/>, 'Moonshot' 'Kimi' <https://www.moonshot.ai/>,
    'OpenAI' 'Codex' subscription endpoints, 'Ollama'
    <https://ollama.com/>, and other 'OpenAI'-compatible endpoints.
    Includes an agent loop with tool use and a 'Model Context
    Protocol' client <https://modelcontextprotocol.io/>. API design is
    derived from the 'ellmer' package, reimplemented with only base R,
    'curl', 'jsonlite', and 'tinyoauth'.
License: MIT + file LICENSE
URL: https://github.com/cornball-ai/llm.api
BugReports: https://github.com/cornball-ai/llm.api/issues
Encoding: UTF-8
Imports: curl, jsonlite, tinyoauth (>= 0.1.1)
Suggests: tinytest
NeedsCompilation: no
Packaged: 2026-08-04 19:41:14 UTC; troy
Author: Troy Hernandez [aut, cre] (ORCID:
    <https://orcid.org/0009-0005-4248-604X>),
  cornball.ai [cph],
  ellmer team [cph] (API design derived from ellmer)
Maintainer: Troy Hernandez <troy@cornball.ai>
Depends: R (>= 3.5.0)
Repository: CRAN
Date/Publication: 2026-08-04 21:50:02 UTC
Built: R 4.6.0; ; 2026-08-04 23:03:43 UTC; unix

# tinyrox says don't edit this manually, but it can't stop you!

export(agent)
export(chat)
export(chat_claude)
export(chat_ollama)
export(chat_openai)
export(chat_session)
export(chat_session_anthropic)
export(chat_session_ollama)
export(chat_session_openai)
export(create_agent)
export(history_count_tool_calls)
export(history_tool_calls)
export(list_ollama_models)
export(llm_base)
export(llm_key)
export(mcp_call)
export(mcp_close)
export(mcp_connect)
export(mcp_start)
export(mcp_tools)
export(mcp_tools_for_api)
export(mcp_tools_for_claude)
export(prices_snapshot_date)
export(prices_snapshot_stale)
export(provider_default_model)
export(usage_cost)

S3method(print,mcp_connection)

# llm.api

Minimal-dependency LLM chat interface. Part of [cornyverse](https://github.com/cornball-ai).

## Exports

| Function | Purpose |
|----------|---------|
| `chat(prompt, model)` | Chat with any LLM |
| `chat_openai(prompt)` | OpenAI GPT models |
| `chat_openai_codex(prompt)` | OpenAI Codex via ChatGPT subscription auth |
| `chat_claude(prompt)` | Anthropic Claude (API key) |
| `chat_claude_oauth(prompt)` | Claude on a Claude subscription (OAuth) |
| `chat_ollama(prompt)` | Local Ollama server |
| `list_ollama_models()` | List Ollama models |
| `llm_base(url)` | Set API endpoint |
| `llm_key(key)` | Set API key |

## Providers

- **openai**: GPT-4o, GPT-4o-mini, o1, o3
- **anthropic**: Claude (Sonnet, Haiku) via an API key
- **anthropic_claude**: Claude on a Claude Pro/Max subscription (OAuth, no API key)
- **moonshot**: Kimi K2 via Moonshot's OpenAI-compatible API
- **openai_codex**: Codex Responses via ChatGPT subscription OAuth
- **ollama**: Llama 3.2, Mistral, Gemma, Phi, Qwen (via server)
- **openai_compatible**: any OpenAI-compatible endpoint (OpenRouter, DeepSeek, a local proxy, a corporate gateway) via a custom base URL

## Usage

```r
# Auto-detect provider from model
chat("Hello", model = "gpt-5.4-mini")
chat("Hello", model = "claude-3-5-sonnet-latest")
chat("Hello", model = "kimi-k2.5")

# Use convenience wrappers
chat_ollama("What is R?")
chat_claude("Explain machine learning")

# Explicit Moonshot/Kimi provider
chat("Write a fast parser in R", provider = "moonshot", model = "kimi-k2.5")

# ChatGPT subscription-backed Codex provider (log in once; see below)
chat_openai_codex("Write a small R function")
chat("Refactor this loop", provider = "openai_codex", model = "gpt-5.5")

# Claude subscription-backed provider (log in once; see below)
chat_claude_oauth("Write a haiku about R")
chat("Refactor this loop", provider = "anthropic_claude")

# Any OpenAI-compatible endpoint via a custom base URL (see below)
llm_base("https://openrouter.ai/api/v1")
chat("Refactor this loop", provider = "openai_compatible",
     model = "meta-llama/llama-3-70b-instruct")

# Provider-native web search (the model searches on its own when useful)
chat("What changed in the latest R release?", web_search = TRUE)

# Conversation history
result <- chat("Hi, I'm Troy")
chat("What's my name?", history = result$history)

# Streaming
chat("Write a story", stream = TRUE)
```

Set `MOONSHOT_API_KEY` to use Moonshot/Kimi without overriding your
OpenAI credentials.

### OpenAI Codex (ChatGPT subscription)

The `openai_codex` provider talks to Codex using your ChatGPT
subscription instead of an API key. Authentication is a one-time device
login; the token is cached and refreshed by
[tinyoauth](https://github.com/cornball-ai/tinyoauth), so you log in
once and it persists across R sessions.

```r
# One-time: device-code login. Prints a URL + code to authorize in a
# browser. The token is cached under tools::R_user_dir("tinyoauth").
openai_codex_login()

# Thereafter, just use the provider; credentials come from the cache
# and refresh automatically.
chat_openai_codex("Write a small R function")
chat("Refactor this", provider = "openai_codex", model = "gpt-5.5")
```

Models: `gpt-5.5` (default), `gpt-5.4`, `gpt-5.4-mini`,
`gpt-5.3-codex-spark`.

To use an externally-obtained token instead of logging in, set
`OPENAI_CODEX_ACCESS_TOKEN` (and optionally `OPENAI_CODEX_ACCOUNT_ID`);
these override the cache.

### Claude (Claude subscription)

The `anthropic_claude` provider drives Claude on a Claude Pro/Max
subscription instead of an API key, the same OAuth pattern as
`openai_codex`. The token is cached and refreshed by
[tinyoauth](https://github.com/cornball-ai/tinyoauth), so you log in
once and it persists across R sessions.

```r
# One-time: prints an authorization URL. Approve it and paste the code back.
claude_oauth_login()

# Thereafter, just use the provider; credentials come from the cache
# and refresh automatically.
chat_claude_oauth("Write a haiku about R")
chat("Refactor this", provider = "anthropic_claude")
```

To use an externally-obtained token instead of logging in, set
`ANTHROPIC_CLAUDE_ACCESS_TOKEN`; it overrides the cache.

### OpenAI-compatible endpoints

The `openai_compatible` provider targets any endpoint that speaks the
OpenAI chat-completions format: OpenRouter, DeepSeek, DeepInfra, a local
proxy, or a corporate gateway. Set the endpoint with `llm_base()` or the
`OPENAI_COMPATIBLE_BASE_URL` environment variable; `/chat/completions` is
appended, so include whatever `/v1` prefix the gateway expects. The model
id is passed through untouched and is required (there is no default).

```r
llm_base("https://openrouter.ai/api/v1")
chat("Write a haiku about R",
     provider = "openai_compatible",
     model = "meta-llama/llama-3-70b-instruct")
```

The API key comes from `OPENAI_COMPATIBLE_API_KEY`, falling back to
`OPENAI_API_KEY`; a keyless internal gateway (no `Authorization` header)
works with neither set. A missing base URL or model fails fast with setup
instructions rather than a curl error.

## Web search

`chat()` and `agent()` take a `web_search` argument (`FALSE` by
default, `TRUE`, or a list of `allowed_domains` / `user_location`
options) that runs the model's own server-side search rather than a
separate search API. The model decides when to search; the result
carries `citations` and `searches`.

It's wired for the hosted providers: `openai` / `openai_codex` (the
Responses `web_search` tool), `anthropic` / `anthropic_claude` (the
Messages `web_search` tool), and `moonshot` (its `$web_search`
builtin). Other providers ignore it with a warning.

```r
res <- chat("What changed in the latest R release?", web_search = TRUE)
res$citations
```

## Dependencies

`curl`, `jsonlite`, and
[tinyoauth](https://github.com/cornball-ai/tinyoauth) (for Codex device
login and token caching). No tidyverse, no compiled code.

# llm.api

Minimal-dependency LLM chat interface. Part of [cornyverse](https://github.com/cornball-ai).

## Exports

| Function | Purpose |
|----------|---------|
| `chat(prompt, model)` | Chat with any LLM |
| `chat_openai(prompt)` | OpenAI GPT models |
| `chat_claude(prompt)` | Anthropic Claude models |
| `chat_ollama(prompt)` | Local Ollama server |
| `list_ollama_models()` | List Ollama models |
| `llm_base(url)` | Set API endpoint |
| `llm_key(key)` | Set API key |

## Providers

- **openai**: GPT-4o, GPT-4o-mini, o1, o3
- **anthropic**: Claude 3.5 Sonnet, Claude 3.5 Haiku
- **moonshot**: Kimi K2 via Moonshot's OpenAI-compatible API
- **ollama**: Llama 3.2, Mistral, Gemma, Phi, Qwen (via server)

## Usage

```r
# Auto-detect provider from model
chat("Hello", model = "gpt-5.4-mini")
chat("Hello", model = "claude-3-5-sonnet-latest")
chat("Hello", model = "kimi-k2.5")

# Use convenience wrappers
chat_ollama("What is R?")
chat_claude("Explain machine learning")

# Explicit Moonshot/Kimi provider
chat("Write a fast parser in R", provider = "moonshot", model = "kimi-k2.5")

# Conversation history
result <- chat("Hi, I'm Troy")
chat("What's my name?", history = result$history)

# Streaming
chat("Write a story", stream = TRUE)
```

Set `MOONSHOT_API_KEY` to use Moonshot/Kimi without overriding your
OpenAI credentials.

## Dependencies

Only `curl` and `jsonlite`. No tidyverse, no compiled code.

# Provider configurations

#' Detect provider from model name or config
#' @noRd
.detect_provider <- function(model) {
    base <- .get_base()

    # Check base URL first
    if (!is.null(base)) {
        if (grepl("moonshot|kimi", base, ignore.case = TRUE)) {
            return("moonshot")
        }
        if (grepl("anthropic", base, ignore.case = TRUE)) {
            return("anthropic")
        }
        if (grepl("openai", base, ignore.case = TRUE)) {
            return("openai")
        }
        if (grepl("localhost|127\\.0\\.0\\.1", base)) {
            return("ollama")
        }
    }

    # Check model name
    if (!is.null(model)) {
        if (grepl("^claude", model, ignore.case = TRUE)) {
            return("anthropic")
        }
        if (grepl("^kimi|^moonshot-v1", model, ignore.case = TRUE)) {
            return("moonshot")
        }
        if (grepl("^(llama|mistral|phi|qwen|gemma|codellama|deepseek)", model,
                  ignore.case = TRUE)) {
            return("ollama")
        }
        # Check if model is installed in Ollama
        ollama_models <- tryCatch(
                                  list_ollama_models()$name,
                                  error = function(e) character(0)
        )
        if (model %in% ollama_models) {
            return("ollama")
        }
        if (grepl("^gpt|^o1|^o3", model, ignore.case = TRUE)) {
            return("openai")
        }
    }

    # Default to OpenAI-compatible
    "openai"
}

#' Get provider configuration
#' @noRd
.get_provider_config <- function(provider) {
    base <- .get_base()
    key <- .get_key()

    switch(provider,
           openai = list(provider = "openai",
                         base_url = base %||% "https://api.openai.com",
                         chat_path = "/v1/chat/completions",
                         api_key = .get_key("openai"),
                         default_model = "gpt-5.4-mini"),
           anthropic = list(
                            provider = "anthropic",
                            base_url = base %||% "https://api.anthropic.com",
                            chat_path = "/v1/messages",
                            api_key = .get_key("anthropic"),
                            default_model = "claude-sonnet-4-6"
        ),
           moonshot = list(
                           provider = "moonshot",
                           base_url = base %||% "https://api.moonshot.ai",
                           chat_path = "/v1/chat/completions",
                           api_key = .get_key("moonshot"),
                           default_model = "kimi-k2.5"
        ),
           ollama = list(
                         provider = "ollama",
                         base_url = base %||% "http://localhost:11434",
                         chat_path = "/v1/chat/completions",
                         api_key = NULL,
                         default_model = "qwen3.5:9b"
        ),
           stop("Unknown provider: ", provider, call. = FALSE)
    )
}

#' Default model for a provider
#'
#' Returns the model name `chat()` falls back to when the caller
#' doesn't specify one. Useful for client code that wants to display
#' the resolved model upfront (e.g., in a status line) without
#' duplicating the lookup table.
#'
#' @param provider Character. One of `"openai"`, `"anthropic"`,
#'   `"moonshot"`, `"ollama"`.
#' @return Character. The default model id for that provider.
#' @export
#' @examples
#' provider_default_model("anthropic")
#' provider_default_model("moonshot")
provider_default_model <- function(provider) {
    .get_provider_config(provider)$default_model
}

#' Chat with OpenAI
#'
#' Convenience wrapper for 'OpenAI' models.
#'
#' @inheritParams chat
#' @return The assistant's response as a character string, or a list when
#'   \code{history} is in use. See \code{\link{chat}} for details.
#' @export
#' @examples
#' \dontrun{
#' chat_openai("Explain quantum computing")
#' chat_openai("Write a haiku", model = "gpt-5.4-mini")
#' }
chat_openai <- function(prompt, model = "gpt-5.4-mini", ...) {
    chat(prompt, model = model, provider = "openai", ...)
}

#' Chat with Anthropic Claude
#'
#' Convenience wrapper for 'Anthropic' 'Claude' models.
#'
#' @inheritParams chat
#' @return The assistant's response as a character string, or a list when
#'   \code{history} is in use. See \code{\link{chat}} for details.
#' @export
#' @examples
#' \dontrun{
#' chat_claude("Explain the theory of relativity")
#' chat_claude("Write a poem", model = "claude-haiku-4-5")
#' }
chat_claude <- function(prompt, model = "claude-sonnet-4-6", ...) {
    chat(prompt, model = model, provider = "anthropic", ...)
}

#' Chat with Ollama
#'
#' Convenience wrapper for local 'Ollama' models.
#'
#' @inheritParams chat
#' @return The assistant's response as a character string, or a list when
#'   \code{history} is in use. See \code{\link{chat}} for details.
#' @export
#' @examples
#' \dontrun{
#' chat_ollama("What is machine learning?")
#' chat_ollama("Explain Docker", model = "mistral")
#' }
chat_ollama <- function(prompt, model = "qwen3.5:9b", ...) {
    chat(prompt, model = model, provider = "ollama", ...)
}

#' List Ollama models
#'
#' List all models downloaded in Ollama.
#'
#' @param base_url Character. Ollama server URL (default: http://localhost:11434).
#'
#' @return A data frame with model information (name, size, modified).
#' @export
#' @examples
#' \dontrun{
#' list_ollama_models()
#' }
list_ollama_models <- function(base_url = "http://localhost:11434") {
    url <- paste0(base_url, "/api/tags")

    h <- curl::new_handle()
    resp <- tryCatch(
                     curl::curl_fetch_memory(url, handle = h),
                     error = function(e) {
        stop("Cannot connect to Ollama at ", base_url,
             ". Is Ollama running?", call. = FALSE)
    }
    )

    if (resp$status_code >= 400) {
        stop("Ollama API error (", resp$status_code, ")", call. = FALSE)
    }

    data <- jsonlite::fromJSON(rawToChar(resp$content))

    if (is.null(data$models) || length(data$models) == 0) {
        message("No models found. Use 'ollama pull <model>' to download models.")
        return(invisible(data.frame(name = character(), size = numeric(), modified = character())))
    }

    data.frame(
               name = data$models$name,
               size = round(data$models$size / 1e9, 1),
               modified = as.character(as.POSIXct(data$models$modified_at))
    )
}

# OpenAI Codex provider tests. Offline: network functions are stubbed.

ns <- asNamespace("llm.api")

b64url <- function(x) {
    x <- jsonlite::base64_enc(charToRaw(jsonlite::toJSON(x, auto_unbox = TRUE)))
    x <- sub("=+$", "", x)
    chartr("+/", "-_", x)
}

fake_codex_jwt <- function(account_id = "acct-test") {
    payload <- list(`https://api.openai.com/auth` = list(
        chatgpt_account_id = account_id
    ))
    paste("hdr", b64url(payload), "sig", sep = ".")
}

with_stubbed <- function(name, stub, expr) {
    orig <- get(name, envir = ns, inherits = FALSE)
    assignInNamespace(name, stub, ns = "llm.api")
    tryCatch(force(expr),
             finally = assignInNamespace(name, orig, ns = "llm.api"))
}

old_opts <- options(llm.api_base = NULL, llm.api_key = NULL, llm.api.api_base = NULL, llm.api.api_key = NULL)
on.exit(options(old_opts), add = TRUE)

old_env <- Sys.getenv(c("OPENAI_CODEX_ACCESS_TOKEN", "OPENAI_CODEX_REFRESH_TOKEN",
                        "OPENAI_CODEX_EXPIRES_AT", "OPENAI_CODEX_ACCOUNT_ID"),
                      unset = NA_character_)
on.exit({
    for (name in names(old_env)) {
        if (is.na(old_env[[name]])) {
            Sys.unsetenv(name)
        } else {
            do.call(Sys.setenv, setNames(list(old_env[[name]]), name))
        }
    }
}, add = TRUE)
Sys.unsetenv(c("OPENAI_CODEX_ACCESS_TOKEN", "OPENAI_CODEX_REFRESH_TOKEN",
               "OPENAI_CODEX_EXPIRES_AT", "OPENAI_CODEX_ACCOUNT_ID"))

# Provider config and defaults.
cfg <- llm.api:::.get_provider_config("openai_codex")
expect_equal(cfg$base_url, "https://chatgpt.com/backend-api")
expect_equal(cfg$chat_path, "/codex/responses")
expect_equal(cfg$default_model, "gpt-5.5")
expect_equal(provider_default_model("openai_codex"), "gpt-5.5")

llm_base("https://chatgpt.com/backend-api")
expect_equal(llm.api:::.detect_provider(NULL), "openai_codex")
options(llm.api_base = NULL, llm.api.api_base = NULL)

# Credentials extract account id from a JWT-shaped access token.
creds <- openai_codex_credentials(access_token = fake_codex_jwt("acct-123"))
headers <- creds()
expect_equal(headers$Authorization, paste("Bearer", fake_codex_jwt("acct-123")))
expect_equal(headers$`chatgpt-account-id`, "acct-123")

# Responses body shape: instructions, Responses input, streaming, and tools.
body <- llm.api:::.openai_codex_body(
    messages = list(list(role = "system", content = "sys"),
                    list(role = "user", content = "hi")),
    tools = list(list(type = "function", name = "add",
                      description = "Add", parameters = list(type = "object"))),
    system = NULL,
    model = "gpt-5.5",
    reasoning_effort = "low"
)
expect_equal(body$instructions, "sys")
expect_true(body$stream)
expect_equal(body$input[[1L]]$role, "user")
expect_equal(body$input[[1L]]$content[[1L]]$type, "input_text")
expect_equal(body$tools[[1L]]$name, "add")
expect_equal(body$tool_choice, "auto")
expect_equal(body$reasoning$effort, "low")

# Streaming event merge reconstructs function-call arguments and final output.
merged <- NULL
chunks <- list(
    list(type = "response.output_item.added", output_index = 0,
         item = list(type = "function_call", id = "fc_1", call_id = "call_1",
                     name = "add", arguments = "")),
    list(type = "response.function_call_arguments.delta", output_index = 0,
         delta = "{\"a\":"),
    list(type = "response.function_call_arguments.done", output_index = 0,
         arguments = "{\"a\":1}"),
    list(type = "response.done", response = list(output = list(), usage = NULL))
)
for (chunk in chunks) {
    merged <- llm.api:::.openai_codex_merge_chunk(merged, chunk)
}
expect_equal(merged$output[[1L]]$arguments, "{\"a\":1}")

# Agent loop: first Codex response asks for a tool, second returns text.
call_count <- 0L
captured_bodies <- list()
stub_sse <- function(url, body, headers) {
    call_count <<- call_count + 1L
    captured_bodies[[call_count]] <<- body
    if (call_count == 1L) {
        list(
             output = list(list(type = "function_call", id = "fc_1",
                                call_id = "call_1", name = "add",
                                arguments = "{\"a\":6,\"b\":7}")),
             usage = list(input_tokens = 10L, output_tokens = 5L,
                          input_tokens_details = list(cached_tokens = 0L))
        )
    } else {
        list(
             output = list(list(type = "message", content = list(
                 list(type = "output_text", text = "tool sum is 13")
             ))),
             usage = list(input_tokens = 20L, output_tokens = 3L,
                          input_tokens_details = list(cached_tokens = 0L))
        )
    }
}

local({
    tool_calls <- 0L
    result <- with_stubbed(".openai_codex_post_sse", stub_sse, {
        llm.api::agent(
            "add numbers",
            provider = "openai_codex",
            model = "gpt-5.5",
            credentials = creds,
            tools = list(list(name = "add", description = "Add",
                              input_schema = list(type = "object"))),
            tool_handler = function(name, args) {
                tool_calls <<- tool_calls + 1L
                as.character(args$a + args$b)
            },
            verbose = FALSE
        )
    })
    expect_equal(result$content, "tool sum is 13")
    expect_equal(tool_calls, 1L)
    expect_equal(call_count, 2L)
    expect_equal(captured_bodies[[2L]]$input[[3L]]$type, "function_call_output")
    expect_equal(captured_bodies[[2L]]$input[[3L]]$call_id, "call_1")
    expect_true(result$usage$cost > 0)
})

# chat() dispatches to the Codex helper and attaches cost.
chat_stub <- function(body, config, stream) {
    list(content = "ok", thinking = NULL, finish_reason = NULL,
         usage = list(prompt_tokens = 10L, completion_tokens = 2L,
                      prompt_tokens_details = list(cached_tokens = 0L)))
}
local({
    result <- with_stubbed(".chat_openai_codex", chat_stub, {
        llm.api::chat("hi", provider = "openai_codex", model = "gpt-5.5",
                      credentials = creds)
    })
    expect_equal(result$content, "ok")
    expect_true(result$usage$cost > 0)
})

# The ChatGPT Codex backend rejects output-token caps (both max_tokens and
# max_output_tokens 400 as unsupported), so the body builder drops them. A
# caller's generic max_tokens must not reach /codex/responses.
mk_user <- function() list(list(role = "user", content = "hi"))

# Supplying a token cap warns once per session, then stays quiet (not per turn).
# (Grab the env into a local and mutate it; assigning through `:::` would try to
# rebind the locked namespace binding.)
codex_state <- ns$.codex_state
codex_state$warned_token_cap <- NULL
expect_warning(
    llm.api:::.openai_codex_body(mk_user(), list(), "sys", "gpt-5.5",
                                 max_tokens = 123L),
    "not supported by the Codex backend")

# Direct body builder: max_tokens dropped, not forwarded or remapped (and now
# silent, since the once-per-session warning already fired above).
b_map <- llm.api:::.openai_codex_body(mk_user(), list(), "sys", "gpt-5.5",
                                      max_tokens = 123L)
expect_null(b_map$max_tokens)
expect_null(b_map$max_output_tokens)
expect_silent(llm.api:::.openai_codex_body(mk_user(), list(), "sys", "gpt-5.5",
                                           max_tokens = 99L))

# An explicitly-passed max_output_tokens is dropped too (also unsupported).
b_keep <- llm.api:::.openai_codex_body(mk_user(), list(), "sys", "gpt-5.5",
                                       max_output_tokens = 456L)
expect_null(b_keep$max_output_tokens)

# Other extra params are still forwarded (drop is scoped to the token caps).
b_other <- llm.api:::.openai_codex_body(mk_user(), list(), "sys", "gpt-5.5",
                                        max_tokens = 5L, temperature = 0.5)
expect_null(b_other$max_tokens)
expect_equal(b_other$temperature, 0.5)

# Captured request body through chat(): no token cap reaches /codex/responses.
capture_sse <- function(target) {
    function(url, body, headers) {
        target$body <- body
        list(output = list(list(type = "message", content = list(
            list(type = "output_text", text = "ok")))),
            usage = list(input_tokens = 1L, output_tokens = 1L,
                         input_tokens_details = list(cached_tokens = 0L)))
    }
}
local({
    cap <- new.env()
    with_stubbed(".openai_codex_post_sse", capture_sse(cap), {
        llm.api::chat("hi", provider = "openai_codex", model = "gpt-5.5",
                      credentials = creds, max_tokens = 77L)
    })
    expect_null(cap$body$max_tokens)
    expect_null(cap$body$max_output_tokens)
})

# Captured request body through agent(): same (both paths share the body
# builder via .openai_codex_request()).
local({
    cap <- new.env()
    with_stubbed(".openai_codex_post_sse", capture_sse(cap), {
        llm.api::agent("go", provider = "openai_codex", model = "gpt-5.5",
                       credentials = creds, verbose = FALSE, max_tokens = 88L)
    })
    expect_null(cap$body$max_tokens)
    expect_null(cap$body$max_output_tokens)
})

# --- provider-native web search (openai_codex) -------------------------------
`%||%` <- function(x, y) if (is.null(x)) y else x
tool_types <- function(tools) {
    if (length(tools) == 0L) character(0)
    else vapply(tools, function(t) t$type %||% "?", character(1))
}

# tool builder: off -> NULL; on -> {type:web_search}; options map to filters
expect_null(ns$.openai_codex_web_search_tool(FALSE))
expect_null(ns$.openai_codex_web_search_tool(NULL))
expect_equal(ns$.openai_codex_web_search_tool(TRUE)$type, "web_search")
wt <- ns$.openai_codex_web_search_tool(list(allowed_domains = c("r-project.org")))
expect_equal(wt$filters$allowed_domains[[1]], "r-project.org")
# unsupported options warn once per session
codex_state$warned_web_search_opts <- NULL
expect_warning(ns$.openai_codex_web_search_tool(list(max_uses = 3)), "ignores")

# body injects the web_search tool and consumes the flag (not forwarded raw)
b <- ns$.openai_codex_body(mk_user(), list(), "sys", "gpt-5.5", web_search = TRUE)
expect_true("web_search" %in% tool_types(b$tools))
expect_null(b$web_search)
boff <- ns$.openai_codex_body(mk_user(), list(), "sys", "gpt-5.5")
expect_false("web_search" %in% tool_types(boff$tools))

# parser: web_search_call -> searches (not a tool call); annotations -> citations
resp <- list(output = list(
    list(type = "web_search_call", status = "completed",
         action = list(type = "search", query = "R version",
                       queries = list("R version"))),
    list(type = "message", content = list(list(type = "output_text",
         text = "R 4.6.0",
         annotations = list(list(type = "url_citation",
                                 url = "https://www.r-project.org", title = "R")))))))
parsed <- ns$.openai_codex_parse_response(resp)
expect_equal(parsed$text, "R 4.6.0")
expect_equal(length(parsed$tool_calls), 0L)
expect_equal(length(parsed$searches), 1L)
expect_equal(parsed$searches[[1]]$query, "R version")
expect_equal(length(parsed$citations), 1L)
expect_equal(parsed$citations[[1]]$url, "https://www.r-project.org")

# captured body through chat(): the web_search tool reaches /codex/responses
local({
    cap <- new.env()
    with_stubbed(".openai_codex_post_sse", capture_sse(cap), {
        llm.api::chat("hi", provider = "openai_codex", model = "gpt-5.5",
                      credentials = creds, web_search = TRUE)
    })
    expect_true("web_search" %in% tool_types(cap$body$tools))
})

# guard: non-codex provider warns and ignores web_search
expect_warning(
    llm.api::chat("hi", provider = "ollama", model = "x", web_search = TRUE),
    "not yet supported")

# Test provider detection and configuration

# --- Setup: save and restore options ---
old_opts <- options(llm.api.api_base = NULL, llm.api.api_key = NULL)
on.exit(options(old_opts), add = TRUE)

# --- .detect_provider() from model name ---

# OpenAI models
expect_equal(llm.api:::.detect_provider("gpt-4o"), "openai")
expect_equal(llm.api:::.detect_provider("gpt-4o-mini"), "openai")
expect_equal(llm.api:::.detect_provider("o1-preview"), "openai")

# Anthropic models
expect_equal(llm.api:::.detect_provider("claude-3-5-sonnet-latest"), "anthropic")
expect_equal(llm.api:::.detect_provider("claude-3-opus"), "anthropic")

# Moonshot / Kimi models
expect_equal(llm.api:::.detect_provider("kimi-k2"), "moonshot")
expect_equal(llm.api:::.detect_provider("kimi-k2-turbo-preview"), "moonshot")
expect_equal(llm.api:::.detect_provider("moonshot-v1-8k"), "moonshot")

# Ollama models
expect_equal(llm.api:::.detect_provider("llama3.2"), "ollama")
expect_equal(llm.api:::.detect_provider("mistral"), "ollama")
expect_equal(llm.api:::.detect_provider("phi3"), "ollama")
expect_equal(llm.api:::.detect_provider("qwen2"), "ollama")

# Default to openai for unknown
expect_equal(llm.api:::.detect_provider("unknown-model"), "openai")
expect_equal(llm.api:::.detect_provider(NULL), "openai")

# --- .detect_provider() from base URL ---

llm_base("http://localhost:11434")
expect_equal(llm.api:::.detect_provider(NULL), "ollama")

llm_base("https://api.anthropic.com")
expect_equal(llm.api:::.detect_provider(NULL), "anthropic")

llm_base("https://api.openai.com")
expect_equal(llm.api:::.detect_provider(NULL), "openai")

llm_base("https://api.moonshot.ai")
expect_equal(llm.api:::.detect_provider(NULL), "moonshot")

# Reset
options(llm.api.api_base = NULL)

# --- .get_provider_config() ---

# OpenAI config
cfg <- llm.api:::.get_provider_config("openai")
expect_equal(cfg$base_url, "https://api.openai.com")
expect_equal(cfg$chat_path, "/v1/chat/completions")
expect_equal(cfg$default_model, "gpt-5.4-mini")

# Anthropic config
cfg <- llm.api:::.get_provider_config("anthropic")
expect_equal(cfg$base_url, "https://api.anthropic.com")
expect_equal(cfg$chat_path, "/v1/messages")
expect_equal(cfg$default_model, "claude-sonnet-4-6")

# Moonshot config
cfg <- llm.api:::.get_provider_config("moonshot")
expect_equal(cfg$base_url, "https://api.moonshot.ai")
expect_equal(cfg$chat_path, "/v1/chat/completions")
expect_equal(cfg$default_model, "kimi-k2.5")

# Ollama config
cfg <- llm.api:::.get_provider_config("ollama")
expect_equal(cfg$base_url, "http://localhost:11434")
expect_equal(cfg$chat_path, "/v1/chat/completions")
expect_equal(cfg$default_model, "qwen3.5:9b")
expect_null(cfg$api_key)

# Unknown provider errors
expect_error(llm.api:::.get_provider_config("unknown"), pattern = "Unknown provider")

# provider_default_model() exposes the default model id for each
# supported provider so client code (e.g., status lines) can resolve it
# upfront without reaching into internals.
expect_equal(provider_default_model("openai"), "gpt-5.4-mini")
expect_equal(provider_default_model("anthropic"), "claude-sonnet-4-6")
expect_equal(provider_default_model("moonshot"), "kimi-k2.5")
expect_equal(provider_default_model("ollama"), "qwen3.5:9b")
expect_error(provider_default_model("nonsense"), pattern = "Unknown provider")

# openai_compatible: generic OpenAI-compatible gateway provider
# (OpenRouter, DeepSeek, corporate proxies). All tests are offline:
# they cover config resolution and the fail-fast guards that run
# before any request is attempted.

# --- Setup: neutralize ambient config so resolution is deterministic ---
old_opts <- options(llm.api_base = NULL, llm.api_key = NULL, llm.api.api_base = NULL, llm.api.api_key = NULL)
old_base_env <- Sys.getenv("OPENAI_COMPATIBLE_BASE_URL", NA)
old_key_env <- Sys.getenv("OPENAI_COMPATIBLE_API_KEY", NA)
old_openai_key <- Sys.getenv("OPENAI_API_KEY", NA)
Sys.unsetenv("OPENAI_COMPATIBLE_BASE_URL")
Sys.unsetenv("OPENAI_COMPATIBLE_API_KEY")
Sys.unsetenv("OPENAI_API_KEY")

# --- .get_provider_config() resolution ---

# Nothing set anywhere: no base URL, no default model, empty key
cfg <- llm.api:::.get_provider_config("openai_compatible")
expect_equal(cfg$provider, "openai_compatible")
expect_null(cfg$base_url)
expect_equal(cfg$chat_path, "/chat/completions")
expect_null(cfg$default_model)
expect_equal(cfg$api_key, "")

# Env var supplies the base; a trailing slash is stripped
Sys.setenv(OPENAI_COMPATIBLE_BASE_URL = "https://openrouter.ai/api/v1/")
cfg <- llm.api:::.get_provider_config("openai_compatible")
expect_equal(cfg$base_url, "https://openrouter.ai/api/v1")

# llm_base() wins over the env var
llm_base("https://gateway.example.com/v1")
cfg <- llm.api:::.get_provider_config("openai_compatible")
expect_equal(cfg$base_url, "https://gateway.example.com/v1")
options(llm.api_base = NULL, llm.api.api_base = NULL)
Sys.unsetenv("OPENAI_COMPATIBLE_BASE_URL")

# --- key resolution: OPENAI_COMPATIBLE_API_KEY first, then OPENAI_API_KEY ---

Sys.setenv(OPENAI_API_KEY = "sk-openai-fallback")
cfg <- llm.api:::.get_provider_config("openai_compatible")
expect_equal(cfg$api_key, "sk-openai-fallback")

Sys.setenv(OPENAI_COMPATIBLE_API_KEY = "sk-gateway")
cfg <- llm.api:::.get_provider_config("openai_compatible")
expect_equal(cfg$api_key, "sk-gateway")

Sys.unsetenv("OPENAI_COMPATIBLE_API_KEY")
Sys.unsetenv("OPENAI_API_KEY")

# --- fail-fast guards, no network ---

# No base URL: chat() and agent() error with setup instructions
expect_error(chat("hi", provider = "openai_compatible", model = "some/model"),
             pattern = "needs a base URL")
expect_error(agent("hi", provider = "openai_compatible", model = "some/model"),
             pattern = "needs a base URL")

# Base set but no model: there is no default to fall back to
llm_base("https://gateway.example.com/v1")
expect_error(chat("hi", provider = "openai_compatible"),
             pattern = "no default model")
expect_error(agent("hi", provider = "openai_compatible"),
             pattern = "no default model")
options(llm.api_base = NULL, llm.api.api_base = NULL)

# provider_default_model() reports the absence of a default as NULL
expect_null(provider_default_model("openai_compatible"))

# The guard is inert for every other provider
expect_null(llm.api:::.check_openai_compatible(
    llm.api:::.get_provider_config("ollama"), NULL))

# --- Cleanup ---
options(old_opts)
if (!is.na(old_base_env)) Sys.setenv(OPENAI_COMPATIBLE_BASE_URL = old_base_env)
if (!is.na(old_key_env)) Sys.setenv(OPENAI_COMPATIBLE_API_KEY = old_key_env)
if (!is.na(old_openai_key)) Sys.setenv(OPENAI_API_KEY = old_openai_key)

# Anthropic Claude subscription (OAuth) provider tests. Offline: credentials are
# exercised with an explicit/env token, no network or tinyoauth cache hit.

ns <- asNamespace("llm.api")

old_env <- Sys.getenv("ANTHROPIC_CLAUDE_ACCESS_TOKEN", unset = NA_character_)
on.exit({
    if (is.na(old_env)) {
        Sys.unsetenv("ANTHROPIC_CLAUDE_ACCESS_TOKEN")
    } else {
        Sys.setenv(ANTHROPIC_CLAUDE_ACCESS_TOKEN = old_env)
    }
}, add = TRUE)
Sys.unsetenv("ANTHROPIC_CLAUDE_ACCESS_TOKEN")

# --- .is_anthropic covers both API-key and OAuth providers ---
expect_true(ns$.is_anthropic("anthropic"))
expect_true(ns$.is_anthropic("anthropic_claude"))
expect_false(ns$.is_anthropic("openai"))

# --- provider config ---
cfg <- llm.api:::.get_provider_config("anthropic_claude")
expect_equal(cfg$provider, "anthropic_claude")
expect_equal(cfg$base_url, "https://api.anthropic.com")
expect_equal(cfg$chat_path, "/v1/messages")
expect_equal(cfg$default_model, "claude-sonnet-4-6")
expect_true(is.function(cfg$credentials))
expect_equal(provider_default_model("anthropic_claude"), "claude-sonnet-4-6")

# --- header selection: API key vs OAuth ---
api_headers <- ns$.anthropic_headers(list(api_key = "sk-test", credentials = NULL))
expect_equal(unname(api_headers[["x-api-key"]]), "sk-test")
expect_true(is.na(api_headers["Authorization"]))
expect_equal(unname(api_headers[["anthropic-version"]]), "2023-06-01")

creds <- anthropic_claude_credentials(access_token = "tok-123")
oauth_headers <- ns$.anthropic_headers(list(api_key = NULL, credentials = creds))
expect_equal(unname(oauth_headers[["Authorization"]]), "Bearer tok-123")
expect_equal(unname(oauth_headers[["anthropic-beta"]]), "oauth-2025-04-20")
expect_true(is.na(oauth_headers["x-api-key"]))

# --- credentials: explicit token, env override, and the headers shape ---
h <- anthropic_claude_credentials(access_token = "tok-abc")()
expect_equal(h$Authorization, "Bearer tok-abc")
expect_equal(h[["anthropic-beta"]], "oauth-2025-04-20")

Sys.setenv(ANTHROPIC_CLAUDE_ACCESS_TOKEN = "env-tok")
expect_equal(anthropic_claude_credentials()()$Authorization, "Bearer env-tok")
Sys.unsetenv("ANTHROPIC_CLAUDE_ACCESS_TOKEN")

# --- subscription usage prices like the API provider (same shape) ---
u <- list(input_tokens = 1000L, output_tokens = 500L)
expect_equal(usage_cost("claude-sonnet-4-6", "anthropic_claude", u),
             usage_cost("claude-sonnet-4-6", "anthropic", u))

# --- regression: every provider stays discoverable via formals() ---
# Downstream packages (e.g. corteza) enumerate providers with
# eval(formals(llm.api::agent)$provider). A scalar default silently drops
# every provider but the first, so the literal choice vector must stay.
all_providers <- c("anthropic", "anthropic_claude", "openai", "moonshot",
                   "openai_codex", "ollama")
expect_true(all(all_providers %in% eval(formals(agent)$provider)))
expect_true(all(all_providers %in% eval(formals(create_agent)$provider)))
expect_true("anthropic_claude" %in% eval(formals(chat)$provider))
expect_true("anthropic_claude" %in% eval(formals(chat_session)$provider))

# --- regression: anthropic_claude is web-search enabled ---
# It shares the Anthropic Messages path, so chat()/agent() must not warn
# and reset web_search = FALSE for it before the request is built.
expect_true("anthropic_claude" %in% ns$.web_search_providers())

# --- regression: subscription OAuth needs the Claude Code identity ---
# Anthropic rejects an OAuth request (429) unless the first system block is
# the Claude Code identity. The API-key path must stay byte-identical.
id <- "You are Claude Code, Anthropic's official CLI for Claude."
# OAuth, no caller system -> identity is the only block.
s1 <- ns$.anthropic_system(NULL, "none", oauth = TRUE)
expect_equal(length(s1), 1L)
expect_equal(s1[[1]]$text, id)
# OAuth, with a caller system -> identity first, caller second.
s2 <- ns$.anthropic_system("Be terse.", "none", oauth = TRUE)
expect_equal(s2[[1]]$text, id)
expect_equal(s2[[2]]$text, "Be terse.")
# OAuth + cache -> cache_control rides the last block.
s3 <- ns$.anthropic_system("x", "5m", oauth = TRUE)
expect_equal(s3[[length(s3)]]$cache_control$type, "ephemeral")
# API-key path unchanged: plain string with no cache, NULL with no system.
expect_equal(ns$.anthropic_system("hello", "none", oauth = FALSE), "hello")
expect_null(ns$.anthropic_system(NULL, "none", oauth = FALSE))

# API Configuration

#' Set LLM API Base URL
#'
#' Stores a base URL in the \code{llm.api.api_base} option, which
#' \code{\link{chat}} uses as the default endpoint.
#'
#' @param url Character. Base URL for the API endpoint.
#' @return The previous value of the option, invisibly.
#' @export
#' @examples
#' old <- llm_base("http://localhost:11434")  # 'Ollama'
#' llm_base(old)  # restore
llm_base <- function(url) {
    old <- getOption("llm.api.api_base")
    options(llm.api.api_base = url)
    invisible(old)
}

#' Set LLM API Key
#'
#' Stores an API key in the \code{llm.api.api_key} option, which
#' \code{\link{chat}} prefers over environment variables.
#'
#' @param key Character. API key for authentication.
#' @return The previous value of the option, invisibly.
#' @export
#' @examples
#' old <- llm_key("sk-not-a-real-key")
#' llm_key(old)  # restore
llm_key <- function(key) {
    old <- getOption("llm.api.api_key")
    options(llm.api.api_key = key)
    invisible(old)
}

#' Get API Base URL
#' @noRd
.get_base <- function() {
    getOption("llm.api.api_base")
}

#' Get API Key
#' @noRd
.get_key <- function(provider = NULL) {
    key <- getOption("llm.api.api_key")
    if (is.null(key) || nchar(key) == 0) {
        env_vars <- switch(provider %||% "",
                           anthropic = c("ANTHROPIC_API_KEY"),
                           openai = c("OPENAI_API_KEY"),
                           moonshot = c("MOONSHOT_API_KEY", "OPENAI_API_KEY"),
                           c("OPENAI_API_KEY", "ANTHROPIC_API_KEY", "MOONSHOT_API_KEY")
        )

        for (env_var in env_vars) {
            key <- Sys.getenv(env_var, "")
            if (nchar(key) > 0) {
                break
            }
        }
    }
    key
}

# Test configuration functions

# --- Setup: save and restore options ---
old_opts <- options(llm.api.api_base = NULL, llm.api.api_key = NULL)
old_env <- Sys.getenv(c("OPENAI_API_KEY", "ANTHROPIC_API_KEY", "MOONSHOT_API_KEY"),
                      unset = "")
on.exit(options(old_opts), add = TRUE)
on.exit(do.call(Sys.setenv, as.list(old_env)), add = TRUE)

# --- llm_base() ---

# Returns previous value invisibly
expect_null(llm_base("http://localhost:11434"))

# Sets the option
expect_equal(getOption("llm.api.api_base"), "http://localhost:11434")

# Returns previous value when called again
expect_equal(llm_base("https://api.openai.com"), "http://localhost:11434")
expect_equal(getOption("llm.api.api_base"), "https://api.openai.com")

# --- llm_key() ---

# Reset first
options(llm.api.api_key = NULL)

# Returns previous value (NULL initially)
expect_null(llm_key("sk-test-key"))

# Sets the option
expect_equal(getOption("llm.api.api_key"), "sk-test-key")

# Returns previous value when called again
expect_equal(llm_key("sk-new-key"), "sk-test-key")

# --- .get_key() ---

options(llm.api.api_key = NULL)
Sys.setenv(MOONSHOT_API_KEY = "moonshot-test-key")
expect_equal(llm.api:::.get_key("moonshot"), "moonshot-test-key")

# Test configuration functions

# --- Setup: save and restore options ---
# Clear both the canonical (0.1.8.1+) and legacy option names so a stale
# value from the environment can't leak into these assertions.
old_opts <- options(llm.api_base = NULL, llm.api_key = NULL,
                    llm.api.api_base = NULL, llm.api.api_key = NULL)
old_env <- Sys.getenv(c("OPENAI_API_KEY", "ANTHROPIC_API_KEY", "MOONSHOT_API_KEY"),
                      unset = "")
on.exit(options(old_opts), add = TRUE)
on.exit(do.call(Sys.setenv, as.list(old_env)), add = TRUE)

# --- llm_base() ---

# Returns previous value invisibly
expect_null(llm_base("http://localhost:11434"))

# Sets the canonical option
expect_equal(getOption("llm.api_base"), "http://localhost:11434")

# Returns previous value when called again
expect_equal(llm_base("https://api.openai.com"), "http://localhost:11434")
expect_equal(getOption("llm.api_base"), "https://api.openai.com")

# --- llm_key() ---

# Reset first
options(llm.api_base = NULL, llm.api_key = NULL)

# Returns previous value (NULL initially)
expect_null(llm_key("sk-test-key"))

# Sets the canonical option
expect_equal(getOption("llm.api_key"), "sk-test-key")

# Returns previous value when called again
expect_equal(llm_key("sk-new-key"), "sk-test-key")

# --- .get_key() ---

options(llm.api_key = NULL)
Sys.setenv(MOONSHOT_API_KEY = "moonshot-test-key")
expect_equal(llm.api:::.get_key("moonshot"), "moonshot-test-key")

# --- legacy option names (pre-0.1.8.1) still read, with a deprecation warning ---

# Base: only the legacy name set -> read it, and warn once.
options(llm.api_base = NULL, llm.api.api_base = "https://legacy.example.com")
# Reset the one-time warning latch so this assertion is deterministic
# regardless of test ordering within the session.
rm(list = ls(envir = llm.api:::.deprecated_opt_warned),
   envir = llm.api:::.deprecated_opt_warned)
expect_warning(base_val <- llm.api:::.get_base(), pattern = "deprecated")
expect_equal(base_val, "https://legacy.example.com")

# Canonical name wins over legacy when both are set (no warning).
options(llm.api_base = "https://canonical.example.com")
expect_equal(llm.api:::.get_base(), "https://canonical.example.com")
options(llm.api_base = NULL, llm.api.api_base = NULL)

# Key: legacy name read through .get_key() when no canonical/env key.
options(llm.api_key = NULL, llm.api.api_key = "sk-legacy")
Sys.setenv(OPENAI_API_KEY = "", ANTHROPIC_API_KEY = "", MOONSHOT_API_KEY = "")
rm(list = ls(envir = llm.api:::.deprecated_opt_warned),
   envir = llm.api:::.deprecated_opt_warned)
expect_warning(key_val <- llm.api:::.get_key("openai"), pattern = "deprecated")
expect_equal(key_val, "sk-legacy")
options(llm.api.api_key = NULL)

# Agentic chat with tool use

#' Chat with tool use (agentic mode)
#'
#' Send a prompt to an LLM with tools. Automatically handles tool calls
#' in a loop until the model responds with text only.
#'
#' @param prompt Character. The user message.
#' @param tools List. Tool definitions (from mcp_tools_for_claude or manual).
#' @param tool_handler Function. Called with (name, args), returns result string.
#' @param system Character. System prompt.
#' @param model Character. Model name.
#' @param provider Character. Provider: "anthropic", "openai", "moonshot",
#'   or "ollama".
#' @param max_turns Integer. Maximum tool-use turns (default: 20).
#' @param verbose Logical. Print tool calls and results.
#' @param history List or NULL. Previous conversation history to continue from.
#' @param history_callback Function or NULL. Called as
#'   \code{history_callback(history)} after each assistant message is
#'   appended and after each tool result is appended. Lets callers
#'   snapshot intermediate state so an interrupt mid-turn doesn't lose
#'   the work that was already done. Errors raised inside the callback
#'   are swallowed so telemetry/snapshotting can't break a turn.
#' @param cache Character. Anthropic prompt caching for the system
#'   message: \code{"none"} (default), \code{"5m"}, or \code{"1h"}
#'   ephemeral TTL. Anthropic-only; warns and degrades to \code{"none"}
#'   for other providers.
#' @param thinking_budget_tokens Integer or NULL. Anthropic extended
#'   thinking budget; must be at least 1024 and less than
#'   \code{max_tokens}. Anthropic-only; ignored with a warning for
#'   other providers.
#' @param ... Additional parameters passed to the API.
#'
#' @return List with final response and conversation history. The
#'   returned \code{$usage} carries cumulative \code{input_tokens},
#'   \code{output_tokens}, \code{total_tokens}, and \code{cost} (USD
#'   scalar, derived from the bundled price snapshot; \code{0} for
#'   Ollama; \code{NA_real_} for models not in the snapshot). It also
#'   carries cumulative cache activity: \code{cache_read_input_tokens}
#'   (Anthropic cache reads plus OpenAI/Moonshot cached prompt tokens),
#'   \code{cache_creation_input_tokens} (total Anthropic cache writes),
#'   and the per-TTL split \code{cache_creation$ephemeral_5m_input_tokens}
#'   / \code{cache_creation$ephemeral_1h_input_tokens}. Passing this
#'   \code{$usage} back to \code{\link{usage_cost}} recomputes the same
#'   \code{cost}.
#' @export
#'
#' @examples
#' \dontrun{
#' # With MCP server
#' conn <- mcp_connect("r", "mcp_server.R")
#' tools <- mcp_tools_for_claude(conn)
#'
#' result <- agent(
#'   "What files are in the current directory?",
#'   tools = tools,
#'   tool_handler = function(name, args) {
#'     mcp_call(conn, name, args)$text
#'   }
#' )
#' }
agent <- function(prompt, tools = list(), tool_handler = NULL, system = NULL,
                  model = NULL,
                  provider = c("anthropic", "openai", "moonshot", "ollama"),
                  max_turns = 20L, verbose = TRUE, history = NULL,
                  history_callback = NULL, cache = c("none", "5m", "1h"),
                  thinking_budget_tokens = NULL, ...) {
    provider <- match.arg(provider)
    cache <- match.arg(cache)

    # Anthropic-only feature opt-ins emit a one-time warning when a
    # non-default value is passed against another provider so the
    # caller knows the request will be silently degraded.
    if (!identical(cache, "none") && !identical(provider, "anthropic")) {
        warning("`cache` is Anthropic-only; ignoring for provider \"",
                provider, "\".", call. = FALSE)
        cache <- "none"
    }
    if (!is.null(thinking_budget_tokens)) {
        # max_tokens flows in via ...; pull it out for validation.
        extra_validate <- list(...)
        .validate_thinking_budget(thinking_budget_tokens,
                                  max_tokens = extra_validate$max_tokens)
        if (!identical(provider, "anthropic")) {
            warning("`thinking_budget_tokens` is Anthropic-only; ignoring ",
                    "for provider \"", provider, "\".", call. = FALSE)
            thinking_budget_tokens <- NULL
        }
    }

    if (is.null(tool_handler) && length(tools) > 0) {
        stop("tool_handler required when tools are provided", call. = FALSE)
    }

    config <- .get_provider_config(provider)

    # Default models with tool support
    if (is.null(model)) {
        model <- switch(provider, anthropic = "claude-sonnet-4-6",
                        openai = "gpt-5.4-mini", moonshot = "kimi-k2.5",
                        ollama = "qwen3.5:9b")
    }

    # Convert tools to provider format
    provider_tools <- .convert_tools(tools, provider)

    # Build initial messages (prepend history if provided)
    if (!is.null(history)) {
        messages <- history
    } else {
        messages <- list()
    }
    messages[[length(messages) + 1]] <- list(role = "user", content = prompt)

    turn <- 0L

    # Track cumulative token usage and cost. Cost is summed per turn:
    # cache token classes are per-response, so pricing each turn and
    # adding is correct and lets a single unpriceable turn propagate to
    # an NA total.
    total_input_tokens <- 0L
    total_output_tokens <- 0L
    total_cache_read <- 0L
    total_cache_write_5m <- 0L
    total_cache_write_1h <- 0L
    total_cost <- 0
    cost_na <- FALSE

    while (turn < max_turns) {
        turn <- turn + 1L

        # Make API request with tools
        response <- switch(provider,
                           anthropic = .agent_anthropic(messages, provider_tools, system, model, config,
                cache = cache,
                thinking_budget_tokens = thinking_budget_tokens, ...),
                           openai = .agent_openai(messages, provider_tools, system, model, config, ...),
                           moonshot = .agent_openai(messages, provider_tools, system, model, config, ...),
                           ollama = .agent_ollama(messages, provider_tools, system, model, config, ...)
        )

        # Accumulate token usage and per-turn cost. Uses `[[` exact
        # matching throughout: `$` would partial-match (e.g.
        # prompt_tokens -> prompt_tokens_details).
        if (!is.null(response$usage)) {
            u <- response$usage
            # Anthropic format
            if (!is.null(u[["input_tokens"]])) {
                total_input_tokens <- total_input_tokens + u[["input_tokens"]]
                total_output_tokens <- total_output_tokens + u[["output_tokens"]]
            }
            # OpenAI/Ollama format
            if (!is.null(u[["prompt_tokens"]])) {
                total_input_tokens <- total_input_tokens + u[["prompt_tokens"]]
                total_output_tokens <- total_output_tokens + u[["completion_tokens"]]
            }
            # Cache token classes via the shared extractor so the
            # per-TTL Anthropic write split is captured (not just the
            # flat total); cache_read covers Anthropic reads, and
            # .openai_cached_tokens adds OpenAI cached prompt tokens.
            ct <- .cache_tokens(u)
            total_cache_read <- total_cache_read + ct$read +
                .openai_cached_tokens(u)
            total_cache_write_5m <- total_cache_write_5m + ct$write_5m
            total_cache_write_1h <- total_cache_write_1h + ct$write_1h
            turn_cost <- usage_cost(model, provider, u)
            if (is.na(turn_cost)) {
                cost_na <- TRUE
            } else {
                total_cost <- total_cost + turn_cost
            }
        }

        # Check if done (no tool calls)
        if (length(response$tool_calls) == 0) {
            # Append final assistant message so caller's history is complete
            messages[[length(messages) + 1]] <- response$assistant_message
            .fire_history_callback(history_callback, messages)
            return(list(
                        content = response$text,
                        model = model,
                        provider = provider,
                        turns = turn,
                        history = messages,
                        usage = list(
                                     input_tokens = total_input_tokens,
                                     output_tokens = total_output_tokens,
                                     total_tokens = total_input_tokens + total_output_tokens,
                                     cache_read_input_tokens = total_cache_read,
                                     cache_creation_input_tokens = total_cache_write_5m + total_cache_write_1h,
                                     cache_creation = list(
                                         ephemeral_5m_input_tokens = total_cache_write_5m,
                                         ephemeral_1h_input_tokens = total_cache_write_1h),
                                     cost = if (cost_na) NA_real_ else total_cost
                    )
                ))
        }

        # Add assistant message (carries the tool_use blocks for this round)
        messages[[length(messages) + 1]] <- response$assistant_message
        .fire_history_callback(history_callback, messages)

        # Process tool calls one at a time and append each result to
        # history as it's produced, firing the callback after each.
        # This means an interrupt mid-batch leaves the completed tools'
        # results in the snapshot the callback received, so the caller
        # can preserve them instead of losing the whole batch.
        for (tc in response$tool_calls) {
            if (verbose) {
                cat(sprintf("\n[Tool: %s]\n", tc$name))
                if (length(tc$arguments) > 0) {
                    cat(sprintf("  Args: %s\n",
                                jsonlite::toJSON(tc$arguments, auto_unbox = TRUE)))
                }
            }

            # Call tool handler
            result <- tryCatch(
                               tool_handler(tc$name, tc$arguments),
                               error = function(e) paste("Error:", e$message)
            )

            if (verbose) {
                display <- if (nchar(result) > 500) {
                    paste0(substr(result, 1, 500), "...")
                } else {
                    result
                }
                cat(sprintf("  Result: %s\n", display))
            }

            messages <- .append_tool_result(
                messages,
                list(id = tc$id, name = tc$name, result = result),
                provider
            )
            .fire_history_callback(history_callback, messages)
        }
    }

    warning("Reached max_turns (", max_turns, ")")
    list(
         content = "[Max turns reached]",
         model = model,
         provider = provider,
         turns = turn,
         history = messages,
         usage = list(
                      input_tokens = total_input_tokens,
                      output_tokens = total_output_tokens,
                      total_tokens = total_input_tokens + total_output_tokens,
                      cache_read_input_tokens = total_cache_read,
                      cache_creation_input_tokens = total_cache_write_5m + total_cache_write_1h,
                      cache_creation = list(
                          ephemeral_5m_input_tokens = total_cache_write_5m,
                          ephemeral_1h_input_tokens = total_cache_write_1h),
                      cost = if (cost_na) NA_real_ else total_cost
        )
    )
}

# Convert tools to provider-specific format
.convert_tools <- function(tools, provider) {
    if (length(tools) == 0) {
        return(list())
    }

    switch(provider,
           anthropic = tools, # Already in Claude format

           openai =,
           moonshot = lapply(tools, function(t) {
        list(
             type = "function",
             `function` = list(name = t$name,
                               description = t$description %||% "",
                               parameters = t$input_schema)
        )
    }),

           ollama = lapply(tools, function(t) {
        list(
             type = "function",
             `function` = list(
                               name = t$name,
                               description = t$description %||% "",
                               parameters = t$input_schema
            )
        )
    })
    )
}

# Anthropic request
.agent_anthropic <- function(messages, tools, system, model, config,
                             cache = "none", thinking_budget_tokens = NULL,
                             ...) {
    url <- paste0(config$base_url, config$chat_path)

    body <- list(model = model, messages = messages, max_tokens = 4096)

    if (!is.null(system)) {
        body$system <- .anthropic_system_with_cache(system, cache)
    }
    if (length(tools) > 0) {
        body$tools <- tools
    }
    if (!is.null(thinking_budget_tokens)) {
        body$thinking <- list(type = "enabled",
                              budget_tokens = as.integer(thinking_budget_tokens))
    }

    extra <- list(...)
    for (name in names(extra)) {
        body[[name]] <- extra[[name]]
    }

    headers <- c("Content-Type" = "application/json",
                 "x-api-key" = config$api_key,
                 "anthropic-version" = "2023-06-01")

    resp <- .post_json(url, body, headers)

    # Parse response
    text_parts <- character()
    tool_calls <- list()

    for (block in resp$content) {
        if (block$type == "text") {
            text_parts <- c(text_parts, block$text)
        } else if (block$type == "tool_use") {
            tool_calls[[length(tool_calls) + 1]] <- list(
                id = block$id,
                name = block$name,
                arguments = block$input
            )
        }
    }

    list(
         text = paste(text_parts, collapse = "\n"),
         tool_calls = tool_calls,
         assistant_message = list(role = "assistant", content = resp$content),
         usage = resp$usage # input_tokens, output_tokens
    )
}

# OpenAI request
.agent_openai <- function(messages, tools, system, model, config, ...) {
    url <- paste0(config$base_url, config$chat_path)

    # Build messages with system
    api_messages <- list()
    if (!is.null(system)) {
        api_messages[[1]] <- list(role = "system", content = system)
    }
    api_messages <- c(api_messages, messages)

    body <- list(model = model, messages = api_messages)

    if (length(tools) > 0) {
        body$tools <- tools
    }

    extra <- list(...)
    # OpenAI deprecated max_tokens in favor of max_completion_tokens
    # and reasoning (o-series) models reject max_tokens entirely. Map
    # for the OpenAI endpoint only; Moonshot, which shares this
    # helper, still expects max_tokens.
    if (identical(config$provider, "openai") &&
        !is.null(extra$max_tokens) &&
        is.null(extra$max_completion_tokens)) {
        extra$max_completion_tokens <- extra$max_tokens
        extra$max_tokens <- NULL
    }
    for (name in names(extra)) {
        body[[name]] <- extra[[name]]
    }

    headers <- c("Content-Type" = "application/json",
                 "Authorization" = paste("Bearer", config$api_key))

    resp <- .post_json(url, body, headers)

    # Parse response
    choice <- resp$choices[[1]]
    msg <- choice$message

    tool_calls <- list()
    if (!is.null(msg$tool_calls)) {
        for (tc in msg$tool_calls) {
            args <- tryCatch(
                             jsonlite::fromJSON(tc$`function`$arguments, simplifyVector = FALSE),
                             error = function(e) list()
            )
            tool_calls[[length(tool_calls) + 1]] <- list(
                id = tc$id,
                name = tc$`function`$name,
                arguments = args
            )
        }
    }

    list(
         text = msg$content %||% "",
         tool_calls = tool_calls,
         assistant_message = msg,
         usage = resp$usage # prompt_tokens, completion_tokens, total_tokens
    )
}

# Ollama request (OpenAI-compatible)
.agent_ollama <- function(messages, tools, system, model, config, ...) {
    url <- paste0(config$base_url, config$chat_path)

    api_messages <- list()
    if (!is.null(system)) {
        api_messages[[1]] <- list(role = "system", content = system)
    }
    api_messages <- c(api_messages, messages)

    body <- list(model = model, messages = api_messages, stream = FALSE)

    if (length(tools) > 0) {
        body$tools <- tools
    }

    extra <- list(...)
    for (name in names(extra)) {
        body[[name]] <- extra[[name]]
    }

    headers <- c("Content-Type" = "application/json")

    resp <- .post_json(url, body, headers)

    # Parse response (OpenAI-compatible format: choices[].message)
    msg <- resp$choices[[1]]$message

    tool_calls <- list()
    if (!is.null(msg$tool_calls)) {
        for (i in seq_along(msg$tool_calls)) {
            tc <- msg$tool_calls[[i]]
            # Parse arguments from JSON string (same as OpenAI)
            args <- tryCatch(
                             jsonlite::fromJSON(tc$`function`$arguments,
                    simplifyVector = FALSE),
                             error = function(e) list()
            )
            # Ollama sometimes omits tc$id; synthesize one and write it back
            # into the assistant message so the corresponding role="tool"
            # result message can reference the same id. Without this the
            # canonical tool_calls list and the on-the-wire history disagree
            # on the call id, which breaks history walks.
            synthesized_id <- tc$id %||% paste0("call_", sample(1e9, 1))
            msg$tool_calls[[i]]$id <- synthesized_id
            tool_calls[[length(tool_calls) + 1]] <- list(
                id = synthesized_id,
                name = tc$`function`$name,
                arguments = args
            )
        }
    }

    list(
         text = msg$content %||% "",
         tool_calls = tool_calls,
         assistant_message = msg,
         usage = resp$usage
    )
}

# Add tool results to message history
.add_tool_results <- function(messages, results, provider) {
    # Backwards-compatible batch wrapper. New code paths should call
    # .append_tool_result directly so the history_callback in agent()
    # can fire between each append.
    for (r in results) {
        messages <- .append_tool_result(messages, r, provider)
    }
    messages
}

# Append a single tool result to history in the provider's expected
# shape. For Anthropic, multiple tool_results for one assistant turn
# share a single trailing user message (extended in place); for the
# OpenAI-family providers each tool result is its own role="tool"
# message.
.append_tool_result <- function(messages, result, provider) {
    switch(provider,
           anthropic = {
        block <- list(type = "tool_result", tool_use_id = result$id,
                      content = result$result)
        last <- length(messages)
        if (last >= 1L &&
            identical(messages[[last]]$role, "user") &&
            is.list(messages[[last]]$content) &&
            length(messages[[last]]$content) > 0L &&
            identical(messages[[last]]$content[[1]]$type, "tool_result")) {
            # Extend the existing batch user message.
            messages[[last]]$content <- c(messages[[last]]$content, list(block))
        } else {
            messages[[length(messages) + 1L]] <- list(role = "user",
                content = list(block))
        }
        messages
    },
           openai =,
           moonshot =,
           ollama = {
        messages[[length(messages) + 1L]] <- list(
            role = "tool",
            tool_call_id = result$id,
            name = result$name,
            content = result$result
        )
        messages
    }
    )
}

# Invoke the user-supplied history callback if any, swallowing errors.
# Callers should pass the current full messages list; the callback
# typically uses it to snapshot intermediate state so an interrupt
# mid-turn doesn't lose completed tool calls.
.fire_history_callback <- function(callback, messages) {
    if (is.null(callback)) {
        return(invisible(NULL))
    }
    tryCatch(callback(messages), error = function(e) NULL)
    invisible(NULL)
}

# Helper: POST JSON request
.post_json <- function(url, body, headers) {
    h <- curl::new_handle()
    curl::handle_setopt(h,
                        customrequest = "POST",
                        postfields = jsonlite::toJSON(body, auto_unbox = TRUE, null = "null")
    )
    curl::handle_setheaders(h, .list = as.list(headers))

    resp <- curl::curl_fetch_memory(url, handle = h)

    if (resp$status_code >= 400) {
        err <- tryCatch(
                        jsonlite::fromJSON(rawToChar(resp$content)),
                        error = function(e) list(error = list(message = rawToChar(resp$content)))
        )
        stop("API error (", resp$status_code, "): ",
             err$error$message %||% "Unknown error", call. = FALSE)
    }

    jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
}

#' Create an agent with MCP servers
#'
#' Convenience function that sets up MCP connections and returns
#' a function for chatting with tools.
#'
#' @param servers Named list of server configs. Each can be:
#'   - `list(port = 7850)` for already-running servers
#'   - `list(command = "r", args = "server.R", port = 7850)` to start and connect
#' @param system Character. Default system prompt.
#' @param model Character. Default model.
#' @param provider Character. Provider: "anthropic", "openai", "moonshot",
#'   or "ollama".
#' @param verbose Logical. Print tool calls.
#'
#' @return A function that takes a prompt and returns a response.
#' @export
#'
#' @examples
#' \dontrun{
#' # Connect to already-running server
#' chat_fn <- create_agent(
#'   servers = list(codeR = list(port = 7850)),
#'   system = "You are a helpful coding assistant."
#' )
#'
#' # Or start server automatically
#' chat_fn <- create_agent(
#'   servers = list(
#'     codeR = list(command = "r", args = "mcp_server.R", port = 7850)
#'   )
#' )
#'
#' result <- chat_fn("List files in current directory")
#' }
create_agent <- function(servers = list(), system = NULL, model = NULL,
                         provider = c("anthropic", "openai", "moonshot", "ollama"),
                         verbose = TRUE) {
    provider <- match.arg(provider)

    # Connect to all servers
    connections <- list()
    for (name in names(servers)) {
        srv <- servers[[name]]

        if (!is.null(srv$command)) {
            # Start server and connect
            connections[[name]] <- mcp_start(command = srv$command,
                args = srv$args, port = srv$port, name = name)
        } else {
            # Connect to existing server
            connections[[name]] <- mcp_connect(
                host = srv$host %||% "localhost",
                port = srv$port,
                name = name
            )
        }
    }

    # Gather all tools (in Claude format - will be converted per-provider)
    all_tools <- list()
    tool_map <- list()

    for (name in names(connections)) {
        conn <- connections[[name]]
        for (tool in conn$tools) {
            all_tools[[length(all_tools) + 1]] <- list(
                name = tool$name,
                description = tool$description %||% "",
                input_schema = tool$inputSchema
            )
            tool_map[[tool$name]] <- conn
        }
    }

    # Create tool handler
    handler <- function(name, args) {
        conn <- tool_map[[name]]
        if (is.null(conn)) {
            return(paste("Unknown tool:", name))
        }
        mcp_call(conn, name, args)$text
    }

    # Return chat function
    function(prompt, ...) {
        agent(
              prompt = prompt,
              tools = all_tools,
              tool_handler = handler,
              system = system,
              model = model,
              provider = provider,
              verbose = verbose,
              ...
        )
    }
}

# Verifies that agent() invokes the history_callback at every history
# mutation point: after each assistant message lands AND after each
# tool result is appended. The point is that a caller can use the
# callback to snapshot intermediate state, so an interrupt mid-turn
# doesn't lose completed tool calls.
#
# Stubs llm.api:::.post_json so the test is offline. Each stubbed
# section restores the original via tryCatch(..., finally = ...) so
# an error inside agent() can't leak the stub into later tests.

ns <- asNamespace("llm.api")
orig_post_json <- get(".post_json", envir = ns, inherits = FALSE)

with_stubbed_post_json <- function(stub, expr) {
    assignInNamespace(".post_json", stub, ns = "llm.api")
    tryCatch(force(expr),
             finally = assignInNamespace(".post_json", orig_post_json,
                                         ns = "llm.api"))
}

# --- Anthropic: fires for assistant + each tool_result block --------
# Two-stage stub: first call returns an assistant message with two
# tool_use blocks; second call returns text-only (end of turn).
anth_call_count <- 0L
anth_stub <- function(url, body, headers) {
    anth_call_count <<- anth_call_count + 1L
    if (anth_call_count == 1L) {
        list(
             content = list(
                            list(type = "text", text = "running two tools"),
                            list(type = "tool_use", id = "tu_1",
                                 name = "echo", input = list(x = 1L)),
                            list(type = "tool_use", id = "tu_2",
                                 name = "echo", input = list(x = 2L))
            ),
             usage = list(input_tokens = 10L, output_tokens = 5L)
        )
    } else {
        list(
             content = list(list(type = "text", text = "all done")),
             usage = list(input_tokens = 12L, output_tokens = 3L)
        )
    }
}

anth_snapshots <- list()
anth_record <- function(messages) {
    anth_snapshots[[length(anth_snapshots) + 1L]] <<- list(
                                                          length = length(messages),
                                                          last_role = messages[[length(messages)]]$role,
                                                          last_content_n =
                                                              if (is.list(messages[[length(messages)]]$content)) {
                                                                  length(messages[[length(messages)]]$content)
                                                              } else {
                                                                  NA_integer_
                                                              }
    )
}

anth_result <- with_stubbed_post_json(anth_stub, llm.api::agent(
                                                                prompt = "go",
                                                                tools = list(list(name = "echo", description = "d",
                                                                                  input_schema = list(type = "object"))),
                                                                tool_handler = function(name, args) sprintf("res-%s", args$x),
                                                                model = "claude-test",
                                                                provider = "anthropic",
                                                                verbose = FALSE,
                                                                history_callback = anth_record
))

expect_equal(anth_result$content, "all done")
expect_equal(anth_call_count, 2L)

# Callback should have fired exactly four times:
#   1. After first assistant message (with 2 tool_uses)
#   2. After first tool_result block appended to a new user message
#   3. After second tool_result block extended that same user message
#   4. After the final text-only assistant message
expect_equal(length(anth_snapshots), 4L)
expect_equal(anth_snapshots[[1L]]$length, 2L)
expect_equal(anth_snapshots[[1L]]$last_role, "assistant")
expect_equal(anth_snapshots[[2L]]$length, 3L)
expect_equal(anth_snapshots[[2L]]$last_role, "user")
expect_equal(anth_snapshots[[2L]]$last_content_n, 1L)
expect_equal(anth_snapshots[[3L]]$length, 3L)
expect_equal(anth_snapshots[[3L]]$last_role, "user")
expect_equal(anth_snapshots[[3L]]$last_content_n, 2L)
expect_equal(anth_snapshots[[4L]]$length, 4L)
expect_equal(anth_snapshots[[4L]]$last_role, "assistant")

# --- OpenAI / Moonshot / Ollama: each tool_result is its own message
# Mirrors the Anthropic test but uses OpenAI's chat-completions
# response shape with a separate role="tool" message per result.
oai_call_count <- 0L
oai_stub <- function(url, body, headers) {
    oai_call_count <<- oai_call_count + 1L
    if (oai_call_count == 1L) {
        list(
             choices = list(list(
                                 message = list(
                                                role = "assistant",
                                                content = "",
                                                tool_calls = list(
                                                                  list(id = "call_1", type = "function",
                                                                       `function` = list(name = "echo",
                                                                                         arguments = "{\"x\":1}")),
                                                                  list(id = "call_2", type = "function",
                                                                       `function` = list(name = "echo",
                                                                                         arguments = "{\"x\":2}"))
                                ),
                                                finish_reason = "tool_calls"
                )
            )),
             usage = list(prompt_tokens = 10L, completion_tokens = 5L)
        )
    } else {
        list(
             choices = list(list(
                                 message = list(role = "assistant", content = "all done"),
                                 finish_reason = "stop"
            )),
             usage = list(prompt_tokens = 12L, completion_tokens = 3L)
        )
    }
}

oai_snapshots <- list()
oai_record <- function(messages) {
    last <- messages[[length(messages)]]
    oai_snapshots[[length(oai_snapshots) + 1L]] <<- list(
                                                         length = length(messages),
                                                         last_role = last$role,
                                                         last_tool_call_id = last$tool_call_id %||% NA_character_
    )
}

oai_result <- with_stubbed_post_json(oai_stub, llm.api::agent(
                                                              prompt = "go",
                                                              tools = list(list(type = "function",
                                                                                `function` = list(name = "echo", description = "d"))),
                                                              tool_handler = function(name, args) sprintf("res-%s", args$x),
                                                              model = "gpt-4o-test",
                                                              provider = "openai",
                                                              verbose = FALSE,
                                                              history_callback = oai_record
))

expect_equal(oai_result$content, "all done")
expect_equal(oai_call_count, 2L)

# Same four fires, but on OpenAI each tool result becomes its OWN
# role="tool" message rather than extending one user message.
#   1. After assistant message (with both tool_calls)
#   2. After first tool message (role="tool", tool_call_id="call_1")
#   3. After second tool message (role="tool", tool_call_id="call_2")
#   4. After final assistant text message
expect_equal(length(oai_snapshots), 4L)
expect_equal(oai_snapshots[[1L]]$length, 2L)
expect_equal(oai_snapshots[[1L]]$last_role, "assistant")
expect_equal(oai_snapshots[[2L]]$length, 3L)
expect_equal(oai_snapshots[[2L]]$last_role, "tool")
expect_equal(oai_snapshots[[2L]]$last_tool_call_id, "call_1")
expect_equal(oai_snapshots[[3L]]$length, 4L)
expect_equal(oai_snapshots[[3L]]$last_role, "tool")
expect_equal(oai_snapshots[[3L]]$last_tool_call_id, "call_2")
expect_equal(oai_snapshots[[4L]]$length, 5L)
expect_equal(oai_snapshots[[4L]]$last_role, "assistant")

# --- Callback errors don't break the turn ---------------------------
anth_call_count <- 0L
boom_result <- with_stubbed_post_json(anth_stub, llm.api::agent(
                                                                prompt = "go",
                                                                tools = list(list(name = "echo", description = "d",
                                                                                  input_schema = list(type = "object"))),
                                                                tool_handler = function(name, args) sprintf("res-%s", args$x),
                                                                model = "claude-test",
                                                                provider = "anthropic",
                                                                verbose = FALSE,
                                                                history_callback = function(messages) {
                                                                    stop("simulated telemetry failure")
                                                                }
))

expect_equal(boom_result$content, "all done")
expect_equal(length(boom_result$history), 4L)

# Verifies that .agent_ollama() writes a synthesized tool-call id back
# into the assistant_message stored in history, not just into the
# canonical $tool_calls return field. Without the writeback the history
# walk pairs assistant.tool_calls[i].id with role="tool".tool_call_id
# and they disagree, breaking history_tool_calls().
#
# Stubs llm.api:::.post_json so the test is offline.

ns <- asNamespace("llm.api")

# Snapshot original .post_json so we can restore it at the end.
orig <- get(".post_json", envir = ns, inherits = FALSE)
on_exit_called <- FALSE

# Stub: return an Ollama-shaped chat response with a tool call that
# OMITS the id field. Some Ollama models don't return an id; the bug
# we're guarding against was that the synthesized id leaked only into
# the canonical tool_calls list.
stub <- function(url, body, headers) {
    list(
        choices = list(
            list(
                message = list(
                    role = "assistant",
                    content = "",
                    tool_calls = list(
                        list(
                            type = "function",
                            # NO id field — that's the case being tested.
                            `function` = list(
                                name = "echo",
                                arguments = "{\"x\":1}"
                            )
                        )
                    )
                ),
                finish_reason = "tool_calls"
            )
        ),
        usage = list(prompt_tokens = 1, completion_tokens = 1)
    )
}

assignInNamespace(".post_json", stub, ns = "llm.api")

result <- llm.api:::.agent_ollama(
    messages = list(list(role = "user", content = "go")),
    tools = list(),
    system = NULL,
    model = "test-model",
    config = list(api_key = "x",
                  base_url = "http://stub",
                  chat_path = "/api/chat")
)

# Restore the real .post_json so other tests don't see the stub.
assignInNamespace(".post_json", orig, ns = "llm.api")

# Canonical tool_calls list got a synthesized id.
expect_equal(length(result$tool_calls), 1L)
synthesized_id <- result$tool_calls[[1]]$id
expect_true(is.character(synthesized_id) && length(synthesized_id) == 1L)
expect_true(nzchar(synthesized_id))

# The fix: the same id is now also on the assistant_message that's
# about to be appended to history. Before this fix, the id field on
# the assistant_message's tool_calls was NULL.
am_calls <- result$assistant_message$tool_calls
expect_equal(length(am_calls), 1L)
expect_equal(am_calls[[1]]$id, synthesized_id)

# Cross-check: history_tool_calls() can pair this call with a result
# message that uses the same id (the agent loop builds these via
# .add_tool_results, so we simulate the result message here).
fake_history <- list(
    list(role = "user", content = "go"),
    result$assistant_message,
    list(role = "tool",
         tool_call_id = synthesized_id,
         name = "echo",
         content = "ok")
)
calls <- llm.api::history_tool_calls(fake_history)
expect_equal(length(calls), 1L)
expect_true(calls[[1]]$completed)
expect_equal(calls[[1]]$result, "ok")

# agent() passes a read-only per-call context snapshot to a tool_handler
# that declares a `context` formal; two-argument handlers are unaffected.
# Offline: stubs llm.api:::.post_json, restoring it via finally.

ns <- asNamespace("llm.api")
orig_post_json <- get(".post_json", envir = ns, inherits = FALSE)
with_stubbed_post_json <- function(stub, expr) {
    assignInNamespace(".post_json", stub, ns = "llm.api")
    tryCatch(force(expr),
             finally = assignInNamespace(".post_json", orig_post_json, ns = "llm.api"))
}

# First response: assistant text + two tool_use blocks. Second: end turn.
call_count <- 0L
stub <- function(url, body, headers) {
    call_count <<- call_count + 1L
    if (call_count == 1L) {
        list(content = list(
                            list(type = "text", text = "running two tools"),
                            list(type = "tool_use", id = "tu_1", name = "echo",
                                 input = list(x = 1L)),
                            list(type = "tool_use", id = "tu_2", name = "echo",
                                 input = list(x = 2L))),
             usage = list(input_tokens = 10L, output_tokens = 5L))
    } else {
        list(content = list(list(type = "text", text = "done")),
             usage = list(input_tokens = 1L, output_tokens = 1L))
    }
}

echo_tools <- list(list(name = "echo", description = "d",
                        input_schema = list(type = "object")))

# --- context-aware handler receives the per-call context ---
seen <- list()
ctx_handler <- function(name, args, context = NULL) {
    seen[[length(seen) + 1L]] <<- context
    sprintf("res-%s", args$x)
}
res <- with_stubbed_post_json(stub, llm.api::agent(
                                                   prompt = "go", tools = echo_tools,
                                                   tool_handler = ctx_handler,
                                                   model = "claude-test", provider = "anthropic",
                                                   verbose = FALSE))

expect_equal(res$content, "done")
expect_equal(length(seen), 2L)
# Both calls share the same turn, assistant text, and call_count; the
# call_index advances 1 -> 2.
expect_equal(seen[[1L]]$assistant_text, "running two tools")
expect_equal(seen[[1L]]$agent_turn, 1L)
expect_equal(seen[[1L]]$call_index, 1L)
expect_equal(seen[[1L]]$call_count, 2L)
expect_equal(seen[[1L]]$provider, "anthropic")
expect_equal(seen[[2L]]$call_index, 2L)
expect_equal(seen[[2L]]$assistant_text, "running two tools")

# --- two-argument handler is called unchanged (no error) ---
call_count <- 0L
res2 <- with_stubbed_post_json(stub, llm.api::agent(
                                                    prompt = "go", tools = echo_tools,
                                                    tool_handler = function(name, args) sprintf("r-%s", args$x),
                                                    model = "claude-test", provider = "anthropic",
                                                    verbose = FALSE))
expect_equal(res2$content, "done")

# --- context arrives by name, even declared after `...` ---
# Positional passing would land context inside `...` and leave it NULL.
call_count <- 0L
dots_seen <- list()
dots_handler <- function(name, args, ..., context = NULL) {
    dots_seen[[length(dots_seen) + 1L]] <<- context
    sprintf("res-%s", args$x)
}
res3 <- with_stubbed_post_json(stub, llm.api::agent(
                                                    prompt = "go", tools = echo_tools,
                                                    tool_handler = dots_handler,
                                                    model = "claude-test", provider = "anthropic",
                                                    verbose = FALSE))
expect_equal(res3$content, "done")
expect_equal(length(dots_seen), 2L)
expect_false(is.null(dots_seen[[1L]]))
expect_equal(dots_seen[[1L]]$assistant_text, "running two tools")
expect_equal(dots_seen[[1L]]$call_count, 2L)

# --- call_index/call_count exclude internally-consumed calls ---
# A Moonshot batch with a server-side $web_search call plus two echo
# calls: only the echoes reach the handler, indexed 1..2 of 2 (not 3).
moon_count <- 0L
moon_stub <- function(url, body, headers) {
    moon_count <<- moon_count + 1L
    if (moon_count == 1L) {
        list(choices = list(list(message = list(
                                               role = "assistant", content = "search then two tools",
                                               tool_calls = list(
                                                                 list(id = "ws_1", type = "function", `function` = list(
                                                                                    name = "$web_search", arguments = "{\"search_id\":\"s1\"}")),
                                                                 list(id = "call_1", type = "function", `function` = list(
                                                                                    name = "echo", arguments = "{\"x\":1}")),
                                                                 list(id = "call_2", type = "function", `function` = list(
                                                                                    name = "echo", arguments = "{\"x\":2}"))),
                                               finish_reason = "tool_calls")),
                            usage = list(prompt_tokens = 10L, completion_tokens = 5L)))
    } else {
        list(choices = list(list(message = list(role = "assistant", content = "done"),
                                 finish_reason = "stop")),
             usage = list(prompt_tokens = 1L, completion_tokens = 1L))
    }
}
moon_seen <- list()
moon_handler <- function(name, args, context = NULL) {
    moon_seen[[length(moon_seen) + 1L]] <<- context
    sprintf("m-%s", args$x)
}
moon_res <- with_stubbed_post_json(moon_stub, llm.api::agent(
                                                            prompt = "go",
                                                            tools = list(list(type = "function",
                                                                              `function` = list(name = "echo", description = "d"))),
                                                            tool_handler = moon_handler,
                                                            model = "kimi-test", provider = "moonshot",
                                                            web_search = TRUE, verbose = FALSE))
expect_equal(moon_res$content, "done")
expect_equal(length(moon_seen), 2L)
expect_equal(moon_seen[[1L]]$call_index, 1L)
expect_equal(moon_seen[[1L]]$call_count, 2L)
expect_equal(moon_seen[[2L]]$call_index, 2L)
expect_equal(moon_seen[[2L]]$call_count, 2L)
expect_equal(moon_seen[[1L]]$provider, "moonshot")

# --- regression: anthropic_claude (subscription OAuth) drives the anthropic
# wire, so tool results are appended and the messages array stays valid on the
# turn after a tool call. Pre-fix, .append_tool_result() didn't match the
# literal "anthropic_claude", returned NULL, and corrupted messages (a 400). ---
local({
    old <- Sys.getenv("ANTHROPIC_CLAUDE_ACCESS_TOKEN", unset = NA)
    Sys.setenv(ANTHROPIC_CLAUDE_ACCESS_TOKEN = "test-token")
    on.exit(if (is.na(old)) Sys.unsetenv("ANTHROPIC_CLAUDE_ACCESS_TOKEN")
            else Sys.setenv(ANTHROPIC_CLAUDE_ACCESS_TOKEN = old), add = TRUE)

    captured <- NULL
    ac_stub <- function(url, body, headers) {
        if (length(body$messages) <= 1L) {
            list(content = list(list(type = "tool_use", id = "t1",
                                     name = "echo", input = list(x = 1L))),
                 usage = list(input_tokens = 5L, output_tokens = 2L))
        } else {
            captured <<- body$messages
            list(content = list(list(type = "text", text = "ok")),
                 usage = list(input_tokens = 1L, output_tokens = 1L))
        }
    }
    ac_res <- with_stubbed_post_json(ac_stub, llm.api::agent(
        prompt = "go", tools = echo_tools,
        tool_handler = function(name, args) "echoed",
        model = "claude-test", provider = "anthropic_claude",
        verbose = FALSE))

    expect_equal(ac_res$content, "ok")
    # Post-tool-result request: messages is a valid array carrying the
    # tool_result, not NULL.
    expect_false(is.null(captured))
    expect_true(length(captured) >= 3L)
    expect_identical(captured[[length(captured)]]$content[[1]]$type,
                     "tool_result")
})

# Standard OpenAI Responses path (provider = "openai" + web_search).
# Offline: the shared SSE poster is stubbed. Asserts that web search routes
# openai through /v1/responses with a bearer key and the web_search tool,
# while a web-search-off openai call stays on chat-completions.

ns <- asNamespace("llm.api")

`%||%` <- function(x, y) if (is.null(x)) y else x
tool_types <- function(tools) {
    if (length(tools) == 0L) {
        character(0)
    } else {
        vapply(tools, function(t) t$type %||% "?", character(1))
    }
}

with_stubbed <- function(name, stub, expr) {
    orig <- get(name, envir = ns, inherits = FALSE)
    assignInNamespace(name, stub, ns = "llm.api")
    tryCatch(force(expr),
             finally = assignInNamespace(name, orig, ns = "llm.api"))
}

# Captures the URL/body/headers handed to the shared SSE poster and returns a
# minimal Responses payload (a message plus one search and one citation).
capture_responses <- function(target) {
    function(url, body, headers) {
        target$url <- url
        target$body <- body
        target$headers <- headers
        list(output = list(
            list(type = "web_search_call", status = "completed",
                 action = list(type = "search", query = "news today",
                               queries = list("news today"))),
            list(type = "message", content = list(list(
                type = "output_text", text = "ok",
                annotations = list(list(type = "url_citation",
                                        url = "https://example.com",
                                        title = "Example")))))),
            usage = list(input_tokens = 5L, output_tokens = 2L,
                         input_tokens_details = list(cached_tokens = 0L)))
    }
}

old_opts <- options(llm.api_base = NULL, llm.api_key = NULL, llm.api.api_base = NULL, llm.api.api_key = NULL)
on.exit(options(old_opts), add = TRUE)
old_key <- Sys.getenv("OPENAI_API_KEY", unset = NA_character_)
on.exit({
    if (is.na(old_key)) {
        Sys.unsetenv("OPENAI_API_KEY")
    } else {
        Sys.setenv(OPENAI_API_KEY = old_key)
    }
}, add = TRUE)
Sys.setenv(OPENAI_API_KEY = "sk-test-openai")

# openai is in the web-search provider set.
expect_true("openai" %in% ns$.web_search_providers())

# Body builder: keeps max_output_tokens (the public endpoint honors caps,
# unlike /codex/responses), injects the web_search tool, and consumes the flag.
mk_user <- function() list(list(role = "user", content = "hi"))
b <- ns$.openai_responses_body(mk_user(), list(), "sys", "gpt-5.4-mini",
                               web_search = TRUE, max_tokens = 99L)
expect_true("web_search" %in% tool_types(b$tools))
expect_null(b$web_search)
expect_equal(b$max_output_tokens, 99L)
expect_null(b$max_tokens)
expect_equal(b$instructions, "sys")
expect_true(isFALSE(b$store))

boff <- ns$.openai_responses_body(mk_user(), list(), "sys", "gpt-5.4-mini")
expect_false("web_search" %in% tool_types(boff$tools))

# chat(): web_search routes to /v1/responses with a bearer key; the response
# carries searches + citations.
local({
    cap <- new.env()
    res <- with_stubbed(".openai_codex_post_sse", capture_responses(cap), {
        llm.api::chat("what's new?", provider = "openai", model = "gpt-5.4-mini",
                      web_search = TRUE)
    })
    expect_equal(cap$url, "https://api.openai.com/v1/responses")
    expect_equal(cap$headers[["Authorization"]], "Bearer sk-test-openai")
    expect_true("web_search" %in% tool_types(cap$body$tools))
    expect_equal(length(res$searches), 1L)
    expect_equal(res$searches[[1]]$query, "news today")
    expect_equal(length(res$citations), 1L)
    expect_equal(res$citations[[1]]$url, "https://example.com")
})

# chat() without web_search stays on chat-completions (Responses poster
# untouched); stub it to fail if hit.
local({
    boom <- function(url, body, headers) stop("should not route to Responses")
    hit <- FALSE
    res <- with_stubbed(".openai_codex_post_sse", boom, {
        with_stubbed(".chat_openai_compatible",
                     function(body, config, stream) {
                         list(content = "cc", usage = NULL)
                     }, {
            llm.api::chat("hi", provider = "openai", model = "gpt-5.4-mini")
        })
    })
    expect_equal(res$content, "cc")
})

# agent(): web_search drives the Responses wire (flat tools, /v1/responses).
local({
    cap <- new.env()
    res <- with_stubbed(".openai_codex_post_sse", capture_responses(cap), {
        llm.api::agent("go", provider = "openai", model = "gpt-5.4-mini",
                       web_search = TRUE, verbose = FALSE)
    })
    expect_equal(cap$url, "https://api.openai.com/v1/responses")
    expect_true("web_search" %in% tool_types(cap$body$tools))
    expect_equal(length(res$searches), 1L)
    expect_equal(length(res$citations), 1L)
})

# agent(): a user tool is converted to the flat Responses shape (name at the
# top level, not nested under `function`).
local({
    cap <- new.env()
    tools <- list(list(name = "noop", description = "d",
                       input_schema = list(type = "object")))
    with_stubbed(".openai_codex_post_sse", capture_responses(cap), {
        llm.api::agent("go", tools = tools,
                       tool_handler = function(n, a) "x",
                       provider = "openai", model = "gpt-5.4-mini",
                       web_search = TRUE, verbose = FALSE)
    })
    user_tool <- Filter(function(t) identical(t$type, "function"), cap$body$tools)
    expect_equal(length(user_tool), 1L)
    expect_equal(user_tool[[1]]$name, "noop")
    expect_null(user_tool[[1]]$`function`)
})

# Moonshot $web_search tests. Offline: the JSON poster is stubbed to simulate
# the builtin-search round-trip (a $web_search tool call, then a final answer).

ns <- asNamespace("llm.api")

`%||%` <- function(x, y) if (is.null(x)) y else x

with_stubbed <- function(name, stub, expr) {
    orig <- get(name, envir = ns, inherits = FALSE)
    assignInNamespace(name, stub, ns = "llm.api")
    tryCatch(force(expr),
             finally = assignInNamespace(name, orig, ns = "llm.api"))
}

old_opts <- options(llm.api_base = NULL, llm.api_key = NULL, llm.api.api_base = NULL, llm.api.api_key = NULL)
on.exit(options(old_opts), add = TRUE)
old_key <- Sys.getenv("MOONSHOT_API_KEY", unset = NA_character_)
on.exit({
    if (is.na(old_key)) {
        Sys.unsetenv("MOONSHOT_API_KEY")
    } else {
        Sys.setenv(MOONSHOT_API_KEY = old_key)
    }
}, add = TRUE)
Sys.setenv(MOONSHOT_API_KEY = "sk-test-moonshot")

# moonshot is in the web-search provider set.
expect_true("moonshot" %in% ns$.web_search_providers())

# tool builder: off -> NULL; on -> the $web_search builtin.
expect_null(ns$.moonshot_web_search_tool(FALSE))
expect_null(ns$.moonshot_web_search_tool(NULL))
wt <- ns$.moonshot_web_search_tool(TRUE)
expect_equal(wt$type, "builtin_function")
expect_equal(wt$`function`$name, "$web_search")

# unsupported options warn once per session. Bind the state env locally first:
# `ns$.moonshot_state$x <- NULL` would try to rebind the locked namespace
# binding, but mutating a local handle to the same env is fine.
moonshot_state <- ns$.moonshot_state
moonshot_state$warned_web_search_opts <- NULL
expect_warning(ns$.moonshot_web_search_tool(list(allowed_domains = "x")),
               "ignores")

# echo coercion: a JSON *string*, not a "json"-classed object (else jsonlite
# would re-embed it as a raw object in the next request body).
echo <- ns$.moonshot_web_search_echo(list(search_result = list(search_id = "s1")))
expect_true(is.character(echo))
expect_false(inherits(echo, "json"))
expect_equal(echo, "{\"search_result\":{\"search_id\":\"s1\"}}")

# A stubbed two-step round-trip: first response asks for a $web_search; second
# (after the echo) returns the final answer. Records the messages it was sent.
mk_poster <- function(target) {
    target$calls <- 0L
    function(url, body, headers) {
        target$calls <- target$calls + 1L
        target$last_messages <- body$messages
        target$last_tools <- body$tools
        if (target$calls == 1L) {
            list(choices = list(list(
                finish_reason = "tool_calls",
                message = list(role = "assistant", content = "",
                    tool_calls = list(list(
                        id = "t-web_search-1", type = "builtin_function",
                        `function` = list(name = "$web_search",
                            arguments = "{\"search_result\":{\"search_id\":\"s1\"}}"))))
            )), usage = list(prompt_tokens = 1L, completion_tokens = 1L))
        } else {
            list(choices = list(list(
                finish_reason = "stop",
                message = list(role = "assistant",
                    content = "Seattle won. (source: example.com)")
            )), usage = list(prompt_tokens = 5L, completion_tokens = 3L))
        }
    }
}

# chat(): drives the echo loop internally and returns the final text + a search
# record (query = NA, since Moonshot doesn't expose it). Two HTTP calls.
local({
    cap <- new.env()
    res <- with_stubbed(".post_json", mk_poster(cap), {
        llm.api::chat("news?", provider = "moonshot", model = "kimi-k2.5",
                      web_search = TRUE)
    })
    expect_equal(cap$calls, 2L)
    expect_equal(res$content, "Seattle won. (source: example.com)")
    expect_equal(length(res$searches), 1L)
    expect_true(is.na(res$searches[[1]]$query))
    expect_equal(length(res$citations), 0L)
    # The builtin tool was sent on each request.
    expect_equal(cap$last_tools[[1]]$`function`$name, "$web_search")
    # The echoed tool result carries the call's arguments verbatim.
    echoed <- cap$last_messages[[length(cap$last_messages)]]
    expect_equal(echoed$role, "tool")
    expect_equal(echoed$name, "$web_search")
    expect_equal(echoed$content, "{\"search_result\":{\"search_id\":\"s1\"}}")
})

# agent(): a user tool plus web search. The $web_search call is intercepted
# (echoed, not routed to tool_handler); the user tool still runs.
local({
    cap <- new.env()
    cap$calls <- 0L
    handler_hits <- new.env()
    handler_hits$names <- character(0)
    handler <- function(name, args) {
        handler_hits$names <- c(handler_hits$names, name)
        "user-result"
    }
    poster <- function(url, body, headers) {
        cap$calls <- cap$calls + 1L
        if (cap$calls == 1L) {
            list(choices = list(list(finish_reason = "tool_calls",
                message = list(role = "assistant", content = "",
                    tool_calls = list(list(id = "t-web_search-1",
                        type = "builtin_function",
                        `function` = list(name = "$web_search",
                            arguments = "{\"search_result\":{\"search_id\":\"s1\"}}")))))),
                usage = list(prompt_tokens = 1L, completion_tokens = 1L))
        } else {
            list(choices = list(list(finish_reason = "stop",
                message = list(role = "assistant", content = "done"))),
                usage = list(prompt_tokens = 2L, completion_tokens = 1L))
        }
    }
    res <- with_stubbed(".post_json", poster, {
        llm.api::agent("go", provider = "moonshot", model = "kimi-k2.5",
                       web_search = TRUE, verbose = FALSE, max_turns = 5)
    })
    expect_equal(res$content, "done")
    expect_equal(length(res$searches), 1L)
    # tool_handler never saw $web_search.
    expect_false("$web_search" %in% handler_hits$names)
})

# guard still fires for a provider with no web search wired.
expect_warning(
    llm.api::chat("hi", provider = "ollama", model = "x", web_search = TRUE),
    "not yet supported")

# Per-call USD cost from token counts.
#
# `.PRICES` is a baked-in snapshot of BerriAI/litellm's
# model_prices_and_context_window.json (see data-raw/prices.R).
# Lookup is offline; CRAN-friendly. Refresh by re-running the snapshot
# script.

# Coerce a possibly-NULL / possibly-NA token count to a numeric scalar,
# treating absent or NA as zero.
#' @noRd
.num0 <- function(x) {
    if (is.null(x) || (length(x) == 1L && is.na(x))) {
        return(0)
    }
    as.numeric(x)
}

#' Look up cost in USD for a single API call.
#'
#' Returns `0` for ollama (local, no API charge), a positive number
#' when the model is in the snapshot, or `NA_real_` when it isn't.
#'
#' `input_tokens` is the *non-cached* prompt token count; cache tokens
#' are passed separately and priced on top. Anthropic cache tokens are
#' priced as published multiples of the base input rate (5-minute write
#' 1.25x, 1-hour write 2x, read 0.1x). For other providers there is no
#' separate write billing; `cache_read` is priced at the snapshot's
#' per-model `cache_read` rate, and the call returns `NA_real_` when
#' cache reads occurred but no such rate is bundled.
#'
#' @param model Character. Model id as sent to the provider.
#' @param provider Character. "anthropic", "openai", "moonshot", or
#'   "ollama".
#' @param input_tokens Integer. Non-cached prompt tokens.
#' @param output_tokens Integer. Completion tokens.
#' @param cache_write_5m,cache_write_1h Integer. Anthropic 5-minute /
#'   1-hour cache-write tokens.
#' @param cache_read Integer. Cache-hit tokens (Anthropic or
#'   OpenAI-compatible).
#' @return Numeric scalar (USD) or `NA_real_`.
#' @noRd
.cost_for <- function(model, provider, input_tokens, output_tokens,
                      cache_write_5m = 0, cache_write_1h = 0, cache_read = 0) {
    if (identical(provider, "ollama")) {
        return(0)
    }
    if (is.null(model) || !nzchar(model)) {
        return(NA_real_)
    }

    prices <- .price_lookup(model, provider)
    if (is.null(prices)) {
        return(NA_real_)
    }

    inp <- .num0(input_tokens)
    out <- .num0(output_tokens)
    w5 <- .num0(cache_write_5m)
    w1 <- .num0(cache_write_1h)
    rd <- .num0(cache_read)

    base <- inp * prices$input + out * prices$output

    if (identical(provider, "anthropic")) {
        return(base +
               w5 * prices$input * 1.25 +
               w1 * prices$input * 2 +
               rd * prices$input * 0.1)
    }

    # OpenAI / Moonshot: no cache-write charge; reads at the snapshot
    # cache_read rate. Unknown rate with reads present -> NA (honest
    # rather than silently billing reads at the full input rate).
    if (rd > 0) {
        if (is.null(prices$cache_read)) {
            return(NA_real_)
        }
        return(base + rd * prices$cache_read)
    }
    base
}

# Anthropic prompt-cache token counts from a usage list, as
# list(write_5m, write_1h, read); zeros when absent. Prefers the
# per-TTL split under `cache_creation`, else treats the flat
# `cache_creation_input_tokens` total as 5-minute writes. Uses `[[`
# throughout: `$` partial-matches, so `cache_creation` would wrongly
# resolve to `cache_creation_input_tokens` in the flat shape.
#' @noRd
.cache_tokens <- function(usage) {
    read <- .num0(usage[["cache_read_input_tokens"]])
    cc <- usage[["cache_creation"]]
    if (is.null(cc)) {
        w5 <- NULL
    } else {
        w5 <- cc[["ephemeral_5m_input_tokens"]]
    }
    if (is.null(cc)) {
        w1 <- NULL
    } else {
        w1 <- cc[["ephemeral_1h_input_tokens"]]
    }
    if (is.null(w5) && is.null(w1)) {
        w5 <- usage[["cache_creation_input_tokens"]]
        w1 <- 0
    }
    list(write_5m = .num0(w5), write_1h = .num0(w1), read = read)
}

# OpenAI-compatible cached prompt-token count. The nested
# `prompt_tokens_details` arrives as a named list under both the
# default and simplifyVector = FALSE JSON parses, so `[[` reads it
# safely either way (and avoids `$` partial matching).
#' @noRd
.openai_cached_tokens <- function(usage) {
    d <- usage[["prompt_tokens_details"]]
    if (is.null(d)) {
        return(0)
    }
    .num0(d[["cached_tokens"]])
}

#' Estimate the USD cost of one call's token usage
#'
#' Computes the offline cost estimate for a usage object, the same
#' value `chat()` and `agent()` attach as `usage$cost`. Reads whichever
#' shape the provider returned (Anthropic's `input_tokens` /
#' `output_tokens`, or the OpenAI-compatible `prompt_tokens` /
#' `completion_tokens`) and accounts for prompt caching: Anthropic
#' cache writes/reads via published multipliers, OpenAI / Moonshot
#' cache hits at the bundled per-model `cache_read` rate. OpenAI cache
#' hits are read from `prompt_tokens_details$cached_tokens` on a raw
#' response, falling back to a flat `cache_read_input_tokens` field as
#' exposed by `agent()` aggregates, so an `agent()$usage` object
#' recomputes to the same cost it already carries.
#'
#' Costs come from the bundled price snapshot, so they are offline,
#' approximate, and may differ from current provider billing. See
#' [prices_snapshot_date()].
#'
#' @param model Character. Model id as sent to the provider.
#' @param provider Character. "anthropic", "openai", "moonshot", or
#'   "ollama".
#' @param usage A usage list as found in `chat()$usage` or
#'   `agent()$usage`.
#' @return Numeric scalar (USD), or `NA_real_` when `usage` is `NULL`,
#'   the model isn't in the snapshot, or cache reads can't be priced.
#' @export
#' @examples
#' \dontrun{
#' r <- chat("hi", model = "claude-sonnet-4-6", cache = "5m")
#' usage_cost("claude-sonnet-4-6", "anthropic", r$usage)
#' }
usage_cost <- function(model, provider, usage) {
    if (is.null(usage)) {
        return(NA_real_)
    }
    out_tokens <- usage[["output_tokens"]] %||% usage[["completion_tokens"]]

    if (identical(provider, "anthropic")) {
        ct <- .cache_tokens(usage)
        return(.cost_for(model, provider,
                         input_tokens = usage[["input_tokens"]],
                         output_tokens = out_tokens,
                         cache_write_5m = ct$write_5m,
                         cache_write_1h = ct$write_1h, cache_read = ct$read))
    }

    # OpenAI-compatible: prompt_tokens includes cached tokens, so split
    # them out and price the cached portion at the cache_read rate. A
    # raw response carries cached reads under prompt_tokens_details;
    # agent() aggregates expose them flat as cache_read_input_tokens, so
    # fall back to that shape.
    prompt <- .num0(usage[["prompt_tokens"]] %||% usage[["input_tokens"]])
    cached <- .openai_cached_tokens(usage)
    if (cached == 0) {
        cached <- .num0(usage[["cache_read_input_tokens"]])
    }
    uncached <- max(prompt - cached, 0)
    .cost_for(model, provider,
              input_tokens = uncached,
              output_tokens = out_tokens,
              cache_read = cached)
}

# Find a model in `.PRICES`. Tries bare model id first, then a
# provider-prefixed form ("openai/gpt-4o", "moonshot/kimi-k2.5") since
# litellm namespaces models that share a basename across providers.
#' @noRd
.price_lookup <- function(model, provider) {
    rec <- .PRICES[[model]]
    if (!is.null(rec)) {
        return(rec)
    }
    if (!is.null(provider) && nzchar(provider)) {
        rec <- .PRICES[[paste0(provider, "/", model)]]
        if (!is.null(rec)) {
            return(rec)
        }
    }
    NULL
}

#' Bundled price-snapshot date
#'
#' `llm.api` estimates `usage$cost` from a model-price table baked into
#' the package at release time. This returns the date that bundled
#' table was generated.
#'
#' It does not contact the network, check for newer prices, or update
#' the installed package. Use the date to display staleness warnings
#' (see [prices_snapshot_stale()]) or to decide when the package
#' maintainer should regenerate the snapshot for a future release.
#'
#' The table is generated from BerriAI/litellm's
#' `model_prices_and_context_window.json`
#' (\url{https://github.com/BerriAI/litellm/blob/main/model_prices_and_context_window.json}).
#' Cost estimates are offline and approximate, and may differ from
#' current provider billing. Cached-input pricing follows each
#' provider's published model: OpenAI
#' (\url{https://platform.openai.com/docs/guides/prompt-caching}),
#' Moonshot (\url{https://www.kimi.com/help/kimi-api/api-pricing}), and
#' Anthropic
#' (\url{https://platform.claude.com/en/docs/build-with-claude/prompt-caching}).
#'
#' @return Character scalar in `YYYY-MM-DD` format.
#' @seealso [prices_snapshot_stale()], [usage_cost()]
#' @export
#' @examples
#' prices_snapshot_date()
prices_snapshot_date <- function() {
    .PRICES_SNAPSHOT_DATE
}

#' Is the bundled price snapshot stale?
#'
#' Convenience wrapper over [prices_snapshot_date()] for staleness
#' alerts, so callers don't repeat the date arithmetic. Offline only;
#' it does not check the network for newer prices.
#'
#' @param max_age_days Numeric. Age threshold in days; default 90.
#' @return `TRUE` when the bundled snapshot is older than
#'   `max_age_days`, otherwise `FALSE`.
#' @seealso [prices_snapshot_date()]
#' @export
#' @examples
#' prices_snapshot_stale()
#' prices_snapshot_stale(max_age_days = 30)
prices_snapshot_stale <- function(max_age_days = 90) {
    age <- as.numeric(Sys.Date() - as.Date(.PRICES_SNAPSHOT_DATE))
    age > max_age_days
}

# Cost lookup against the baked price snapshot. Pure-function,
# offline. The snapshot ships in R/sysdata.rda and is regenerated
# by data-raw/prices.R.

# ---- snapshot is present and dated ----

snap <- llm.api::prices_snapshot_date()
expect_true(is.character(snap) && nzchar(snap),
            info = "prices_snapshot_date() returns a non-empty string")
expect_true(grepl("^[0-9]{4}-[0-9]{2}-[0-9]{2}$", snap),
            info = "snapshot date is ISO YYYY-MM-DD")

# ---- Known models (rates from BerriAI/litellm) ----
# 1000 input + 500 output tokens, costs in USD.

# OpenAI gpt-4o: input 2.5e-06, output 1e-05 -> 0.0075
expect_equal(llm.api:::.cost_for("gpt-4o", "openai", 1000, 500), 0.0075)

# Anthropic claude-sonnet-4-6: input 3e-06, output 1.5e-05 -> 0.0105
expect_equal(llm.api:::.cost_for("claude-sonnet-4-6", "anthropic", 1000, 500),
             0.0105)

# Dated Anthropic id resolves to the same per-token rate.
expect_equal(llm.api:::.cost_for("claude-sonnet-4-20250514", "anthropic", 1000, 500),
             0.0105)

# ---- Ollama short-circuits to zero ----

expect_equal(llm.api:::.cost_for("llama3.2", "ollama", 1000, 500), 0)
expect_equal(llm.api:::.cost_for("anything", "ollama", 0, 0), 0)

# ---- Unknown / null model -> NA_real_ ----

expect_identical(llm.api:::.cost_for("totally-not-a-model-xyz", "openai", 1000, 500),
                 NA_real_)
expect_identical(llm.api:::.cost_for(NULL, "openai", 1000, 500), NA_real_)
expect_identical(llm.api:::.cost_for("", "openai", 1000, 500), NA_real_)

# ---- Provider-prefixed fallback ----
# litellm stores some Moonshot models only under "moonshot/<id>". The
# fallback should find them when bare lookup misses.

if (!is.null(llm.api:::.PRICES[["moonshot/kimi-k2.5"]])) {
    expected <- with(llm.api:::.PRICES[["moonshot/kimi-k2.5"]],
                     input * 1000 + output * 500)
    expect_equal(llm.api:::.cost_for("kimi-k2.5", "moonshot", 1000, 500),
                 expected,
                 info = "moonshot prefix fallback hits moonshot/kimi-k2.5")
}

# ---- NA / NULL token counts treated as zero ----

# Known model with no tokens at all -> 0, not NA.
expect_equal(llm.api:::.cost_for("gpt-4o", "openai", 0, 0), 0)
expect_equal(llm.api:::.cost_for("gpt-4o", "openai", NA, NA), 0)
expect_equal(llm.api:::.cost_for("gpt-4o", "openai", NULL, NULL), 0)

# ---- chat() result wiring: .augment_usage_with_cost ----
# Works against both Anthropic-shaped and OpenAI-shaped usage lists.

augment <- llm.api:::.augment_usage_with_cost

anthropic_usage <- list(input_tokens = 1000, output_tokens = 500)
out_a <- augment(anthropic_usage, "claude-sonnet-4-6", "anthropic")
expect_equal(out_a$cost, 0.0105)
expect_equal(out_a$input_tokens, 1000)
expect_equal(out_a$output_tokens, 500)

openai_usage <- list(prompt_tokens = 1000, completion_tokens = 500,
                     total_tokens = 1500)
out_o <- augment(openai_usage, "gpt-4o", "openai")
expect_equal(out_o$cost, 0.0075)
expect_equal(out_o$prompt_tokens, 1000)
expect_equal(out_o$total_tokens, 1500)

# NULL usage passes through (e.g. streaming returns no usage).
expect_null(augment(NULL, "gpt-4o", "openai"))

# Unknown model returns NA_real_ cost but leaves token fields intact.
out_u <- augment(openai_usage, "not-a-real-model", "openai")
expect_identical(out_u$cost, NA_real_)
expect_equal(out_u$prompt_tokens, 1000)

# ---- Anthropic prompt-cache pricing (multipliers on base input) ----
# claude-sonnet-4-6: input 3e-06, output 1.5e-05. Base 1000/500 = 0.0105.

# 5-minute writes at 1.25x input: +1000*3e-06*1.25 = 0.00375.
expect_equal(llm.api:::.cost_for("claude-sonnet-4-6", "anthropic", 1000, 500,
                                 cache_write_5m = 1000), 0.01425)
# 1-hour writes at 2x input: +1000*3e-06*2 = 0.006.
expect_equal(llm.api:::.cost_for("claude-sonnet-4-6", "anthropic", 1000, 500,
                                 cache_write_1h = 1000), 0.0165)
# Cache reads at 0.1x input: +1000*3e-06*0.1 = 0.0003.
expect_equal(llm.api:::.cost_for("claude-sonnet-4-6", "anthropic", 1000, 500,
                                 cache_read = 1000), 0.0108)
# Four-arg calls are unchanged (cache args default to zero).
expect_equal(llm.api:::.cost_for("claude-sonnet-4-6", "anthropic", 1000, 500),
             0.0105)

# ---- usage_cost(): Anthropic usage shapes ----

# Flat cache_creation total is treated as 5-minute writes.
ant_flat <- list(input_tokens = 1000, output_tokens = 500,
                 cache_creation_input_tokens = 1000,
                 cache_read_input_tokens = 500)
# 0.003 + 0.0075 + 1000*3e-06*1.25 + 500*3e-06*0.1 = 0.0144
expect_equal(llm.api::usage_cost("claude-sonnet-4-6", "anthropic", ant_flat),
             0.0144)

# Per-TTL split prices 5m and 1h writes distinctly.
ant_split <- list(input_tokens = 10, output_tokens = 5,
                  cache_creation = list(ephemeral_5m_input_tokens = 1000,
                                        ephemeral_1h_input_tokens = 2000),
                  cache_read_input_tokens = 0)
# 1e-5*... : 10*3e-06 + 5*1.5e-05 + 1000*3e-06*1.25 + 2000*3e-06*2 = 0.015855
expect_equal(llm.api::usage_cost("claude-sonnet-4-6", "anthropic", ant_split),
             0.015855)

# ---- usage_cost(): OpenAI cached input (data-driven cache_read) ----
# gpt-5.4-mini: input 7.5e-07, output 4.5e-06, cache_read 7.5e-08.
oa <- list(prompt_tokens = 1000, completion_tokens = 500,
           prompt_tokens_details = list(cached_tokens = 400))
# uncached 600*7.5e-07 + 500*4.5e-06 + cached 400*7.5e-08 = 0.00273
expect_equal(llm.api::usage_cost("gpt-5.4-mini", "openai", oa), 0.00273)

# No cached tokens -> plain input pricing.
oa0 <- list(prompt_tokens = 1000, completion_tokens = 500)
expect_equal(llm.api::usage_cost("gpt-5.4-mini", "openai", oa0), 0.003)

# Honest NA: cached tokens present but the model carries no cache_read
# rate in the snapshot -> NA rather than billing reads at full rate.
no_cache <- NULL
for (k in names(llm.api:::.PRICES)) {
    r <- llm.api:::.PRICES[[k]]
    if (identical(r$provider, "openai") && is.null(r$cache_read) &&
        !is.null(r$input) && r$input > 0) {
        no_cache <- k
        break
    }
}
if (!is.null(no_cache)) {
    u_nc <- list(prompt_tokens = 1000, completion_tokens = 100,
                 prompt_tokens_details = list(cached_tokens = 500))
    expect_identical(llm.api::usage_cost(no_cache, "openai", u_nc), NA_real_)
}

# NULL usage -> NA.
expect_identical(llm.api::usage_cost("gpt-5.4-mini", "openai", NULL), NA_real_)

# ---- prices_snapshot_stale() ----
expect_false(llm.api::prices_snapshot_stale(max_age_days = 100000))
expect_true(llm.api::prices_snapshot_stale(max_age_days = -1))

# ---- agent() sums per-turn cost and aggregates cache fields ----
# Stub the Anthropic dispatch with a two-turn run carrying cache tokens.
local({
    ns <- asNamespace("llm.api")
    responses <- list(
        list(text = "",
             assistant_message = list(role = "assistant", content = "t1"),
             tool_calls = list(list(id = "c1", name = "noop", arguments = list())),
             usage = list(input_tokens = 100, output_tokens = 10,
                          cache_creation_input_tokens = 2000,
                          cache_read_input_tokens = 0)),
        list(text = "final",
             assistant_message = list(role = "assistant", content = "final"),
             tool_calls = list(),
             usage = list(input_tokens = 50, output_tokens = 20,
                          cache_read_input_tokens = 2000))
    )
    i <- 0L
    stub <- function(messages, provider_tools, system, model, config, ...) {
        i <<- i + 1L
        responses[[i]]
    }
    orig <- get(".agent_anthropic", envir = ns, inherits = FALSE)
    assignInNamespace(".agent_anthropic", stub, ns = "llm.api")
    res <- tryCatch(
        llm.api::agent(prompt = "go", tools = list(),
                       tool_handler = function(name, args) "ok",
                       provider = "anthropic", model = "claude-sonnet-4-6",
                       verbose = FALSE, max_turns = 5L),
        finally = assignInNamespace(".agent_anthropic", orig, ns = "llm.api"))
    # turn1 0.00795 + turn2 0.00105 = 0.009
    expect_equal(res$usage$cost, 0.009)
    expect_equal(res$usage$input_tokens, 150)
    expect_equal(res$usage$output_tokens, 30)
    expect_equal(res$usage$cache_read_input_tokens, 2000)
    expect_equal(res$usage$cache_creation_input_tokens, 2000)
})

# ---- usage_cost() handles agent()-shaped aggregates ----
# agent() exposes OpenAI cached reads flat as cache_read_input_tokens
# (no prompt_tokens_details). Same price as the raw-response shape.
oa_agg <- list(input_tokens = 1000, output_tokens = 500,
               cache_read_input_tokens = 400)
expect_equal(llm.api::usage_cost("gpt-5.4-mini", "openai", oa_agg), 0.00273)

# agent() aggregates the Anthropic per-TTL write split (not just the
# flat total): a turn with only the nested cache_creation shape must
# still sum into cache_creation_input_tokens and price correctly.
local({
    ns <- asNamespace("llm.api")
    responses <- list(
        list(text = "final",
             assistant_message = list(role = "assistant", content = "x"),
             tool_calls = list(),
             usage = list(input_tokens = 10, output_tokens = 5,
                          cache_creation = list(
                              ephemeral_5m_input_tokens = 1000,
                              ephemeral_1h_input_tokens = 2000),
                          cache_read_input_tokens = 0))
    )
    i <- 0L
    stub <- function(messages, provider_tools, system, model, config, ...) {
        i <<- i + 1L
        responses[[i]]
    }
    orig <- get(".agent_anthropic", envir = ns, inherits = FALSE)
    assignInNamespace(".agent_anthropic", stub, ns = "llm.api")
    res <- tryCatch(
        llm.api::agent(prompt = "go", tools = list(),
                       tool_handler = function(name, args) "ok",
                       provider = "anthropic", model = "claude-sonnet-4-6",
                       verbose = FALSE, max_turns = 3L),
        finally = assignInNamespace(".agent_anthropic", orig, ns = "llm.api"))
    # split writes aggregated: 1000 + 2000
    expect_equal(res$usage$cache_creation_input_tokens, 3000)
    # 10*3e-06 + 5*1.5e-05 + 1000*3e-06*1.25 + 2000*3e-06*2 = 0.015855
    expect_equal(res$usage$cost, 0.015855)
    # agent()$usage preserves the 5m/1h split, so recomputing the cost
    # from the aggregate round-trips to the stored value (a flat-only
    # aggregate would mis-price the 1h writes as 5m).
    expect_equal(res$usage$cache_creation$ephemeral_5m_input_tokens, 1000)
    expect_equal(res$usage$cache_creation$ephemeral_1h_input_tokens, 2000)
    expect_equal(llm.api::usage_cost("claude-sonnet-4-6", "anthropic", res$usage),
                 res$usage$cost)
})

# Core chat functionality

# Validate a thinking budget against Anthropic's documented
# constraints: it must be a positive integer of at least 1024 tokens,
# and (when max_tokens is set) it must leave room for the regular
# completion -- the budget is counted within max_tokens.
.validate_thinking_budget <- function(thinking_budget_tokens,
                                      max_tokens = NULL) {
    if (!is.numeric(thinking_budget_tokens) ||
        length(thinking_budget_tokens) != 1L ||
        is.na(thinking_budget_tokens) ||
        thinking_budget_tokens != as.integer(thinking_budget_tokens)) {
        stop("`thinking_budget_tokens` must be a single integer.",
             call. = FALSE)
    }
    if (thinking_budget_tokens < 1024L) {
        stop("`thinking_budget_tokens` must be at least 1024 ",
             "(Anthropic's documented minimum).", call. = FALSE)
    }
    if (!is.null(max_tokens) &&
        thinking_budget_tokens >= as.integer(max_tokens)) {
        stop("`thinking_budget_tokens` (", thinking_budget_tokens,
             ") must be strictly less than `max_tokens` (",
             max_tokens, "); the thinking budget counts against ",
             "max_tokens and must leave room for the completion.",
             call. = FALSE)
    }
    invisible(TRUE)
}

# Wrap the system message in a cache_control block when caching is
# requested, or pass it through as plain text when cache == "none".
# The "5m" and "1h" values map to Anthropic's ephemeral cache TTLs.
.anthropic_system_with_cache <- function(system_msg, cache) {
    if (identical(cache, "none")) {
        return(system_msg)
    }
    control <- if (identical(cache, "1h")) {
        list(type = "ephemeral", ttl = "1h")
    } else {
        list(type = "ephemeral")
    }
    list(list(type = "text", text = system_msg, cache_control = control))
}

#' Chat with an LLM
#'
#' Send a message to a Large Language Model and get a response.
#'
#' @param prompt Character. The user message to send.
#' @param model Character. Model name (e.g., "gpt-5.4-mini", "claude-sonnet-4-6", "qwen3.5:9b").
#' @param system Character or NULL. System prompt to set context.
#' @param history List or NULL. Previous conversation turns.
#' @param temperature Numeric or NULL. Sampling temperature (0-2).
#' @param max_tokens Integer or NULL. Maximum tokens in response.
#' @param provider Character. Provider: "auto", "openai", "anthropic",
#'   "moonshot", or "ollama".
#' @param stream Logical. Stream the response (prints as it arrives).
#' @param cache Character. Anthropic prompt caching for the system
#'   message: \code{"none"} (default), \code{"5m"}, or \code{"1h"}
#'   ephemeral TTL. Anthropic-only; warns and degrades to \code{"none"}
#'   for other providers.
#' @param thinking_budget_tokens Integer or NULL. Anthropic extended
#'   thinking budget; must be at least 1024 and less than
#'   \code{max_tokens}. Anthropic-only; ignored with a warning for
#'   other providers.
#' @param ... Additional parameters passed to the API.
#'
#' @return A list with:
#'   \item{content}{The assistant's response text}
#'   \item{thinking}{Chain-of-thought from reasoning models, or NULL.
#'     Populated from \code{reasoning_content} (DeepSeek, Moonshot Kimi,
#'     vLLM, SGLang), \code{reasoning} (OpenRouter), or Anthropic
#'     \code{thinking} blocks. Normalized across providers.}
#'   \item{finish_reason}{Why generation stopped. \code{"stop"} on a
#'     normal completion, \code{"length"} when truncated by max_tokens.
#'     A reasoning model that returns empty \code{content} with
#'     \code{finish_reason == "length"} ran out of budget mid-thought;
#'     raise \code{max_tokens}.}
#'   \item{model}{Model used}
#'   \item{usage}{Token usage (if available). When the model is in the
#'     bundled price snapshot, also carries \code{cost} as a USD scalar;
#'     Ollama is treated as free (\code{cost = 0}); unknown models leave
#'     \code{cost = NA_real_}. See \code{\link{prices_snapshot_date}}.}
#'   \item{history}{Updated conversation history}
#'
#' @export
#' @examples
#' \dontrun{
#' # Simple chat
#' chat("What is 2+2?")
#'
#' # With system prompt
#' chat("Explain R", system = "You are a helpful programming tutor.")
#'
#' # Continue conversation
#' result <- chat("Hello")
#' chat("Tell me more", history = result$history)
#' }
chat <- function(prompt, model = NULL, system = NULL, history = NULL,
                 temperature = NULL, max_tokens = NULL,
                 provider = c("auto", "openai", "anthropic", "moonshot", "ollama"),
                 stream = FALSE, cache = c("none", "5m", "1h"),
                 thinking_budget_tokens = NULL, ...) {
    provider <- match.arg(provider)
    cache <- match.arg(cache)

    # Validate the thinking-budget range up front. This is provider-
    # independent input validation and should fail fast, before any
    # provider resolution.
    if (!is.null(thinking_budget_tokens)) {
        .validate_thinking_budget(thinking_budget_tokens, max_tokens)
    }

    # Resolve "auto" to a concrete provider before the Anthropic-only
    # guards below, otherwise they compare against "auto" and wrongly
    # disable cache / thinking_budget_tokens for genuine Anthropic calls.
    if (provider == "auto") {
        provider <- .detect_provider(model)
    }

    # Anthropic-only feature opt-ins emit a one-time warning when a
    # non-default value is passed against another provider so the
    # caller knows the request will be silently degraded.
    if (!identical(cache, "none") && !identical(provider, "anthropic")) {
        warning("`cache` is Anthropic-only; ignoring for provider \"",
                provider, "\".", call. = FALSE)
        cache <- "none"
    }
    if (!is.null(thinking_budget_tokens) && !identical(provider, "anthropic")) {
        warning("`thinking_budget_tokens` is Anthropic-only; ignoring ",
                "for provider \"", provider, "\".", call. = FALSE)
        thinking_budget_tokens <- NULL
    }

    # Get provider config
    config <- .get_provider_config(provider)

    # Set default model if not specified
    if (is.null(model)) {
        model <- config$default_model
    }

    # Build messages array
    messages <- list()

    if (!is.null(system)) {
        messages[[length(messages) + 1]] <- list(role = "system",
            content = system)
    }

    if (!is.null(history)) {
        messages <- c(messages, history)
    }

    messages[[length(messages) + 1]] <- list(role = "user", content = prompt)

    # Build request body
    body <- list(model = model, messages = messages, stream = stream)

    if (!is.null(temperature)) {
        body$temperature <- temperature
    }
    if (!is.null(max_tokens)) {
        body$max_tokens <- max_tokens
    }

    # Add extra params

    extra <- list(...)
    for (name in names(extra)) {
        body[[name]] <- extra[[name]]
    }

    # Make request
    if (provider == "anthropic") {
        result <- .chat_anthropic(body, config, stream,
                                  cache = cache,
                                  thinking_budget_tokens = thinking_budget_tokens)
    } else {
        result <- .chat_openai_compatible(body, config, stream)
    }

    # Build updated history
    new_history <- messages
    new_history[[length(new_history) + 1]] <- list(
        role = "assistant",
        content = result$content
    )

    usage <- .augment_usage_with_cost(result$usage, model, provider)

    list(
         content = result$content,
         thinking = result$thinking,
         finish_reason = result$finish_reason,
         model = model,
         usage = usage,
         history = new_history
    )
}

# Attach a USD cost field to a provider-shaped usage list. Delegates to
# usage_cost(), which reads the Anthropic or OpenAI-compatible token
# shape and accounts for prompt caching. Cost is appended without
# renaming the existing token fields so callers that already
# destructure usage keep working.
#' @noRd
.augment_usage_with_cost <- function(usage, model, provider) {
    if (is.null(usage)) {
        return(usage)
    }
    usage$cost <- usage_cost(model, provider, usage)
    usage
}

#' OpenAI-compatible chat request
#' @noRd
.chat_openai_compatible <- function(body, config, stream) {
    url <- paste0(config$base_url, config$chat_path)

    # OpenAI deprecated max_tokens in favor of max_completion_tokens
    # and reasoning (o-series) models reject max_tokens entirely. Map
    # for the OpenAI endpoint only; Moonshot and Ollama (which share
    # this helper) still expect max_tokens.
    if (identical(config$provider, "openai") &&
        !is.null(body$max_tokens) &&
        is.null(body$max_completion_tokens)) {
        body$max_completion_tokens <- body$max_tokens
        body$max_tokens <- NULL
    }

    headers <- c("Content-Type" = "application/json")

    if (!is.null(config$api_key) && nchar(config$api_key) > 0) {
        headers["Authorization"] <- paste("Bearer", config$api_key)
    }

    h <- curl::new_handle()
    curl::handle_setopt(h, customrequest = "POST",
                        postfields = jsonlite::toJSON(body, auto_unbox = TRUE))
    curl::handle_setheaders(h, .list = as.list(headers))

    if (stream) {
        .stream_response(url, h)
    } else {
        resp <- curl::curl_fetch_memory(url, handle = h)

        if (resp$status_code >= 400) {
            err <- tryCatch(
                            jsonlite::fromJSON(rawToChar(resp$content)),
                            error = function(e) list(error = list(message = rawToChar(resp$content)))
            )
            stop("API error (", resp$status_code, "): ",
                 err$error$message %||% "Unknown error", call. = FALSE)
        }

        data <- jsonlite::fromJSON(rawToChar(resp$content))

        # Handle both list and data.frame formats from jsonlite
        if (is.data.frame(data$choices)) {
            msg <- data$choices$message
            content <- msg$content[1]
            thinking <- msg$reasoning_content[1] %||% msg$reasoning[1]
            finish_reason <- data$choices$finish_reason[1]
        } else {
            msg <- data$choices[[1]]$message
            content <- msg$content
            thinking <- msg$reasoning_content %||% msg$reasoning
            finish_reason <- data$choices[[1]]$finish_reason
        }

        .warn_if_truncated(content, thinking, finish_reason)

        list(
             content = content,
             thinking = thinking,
             finish_reason = finish_reason,
             usage = data$usage
        )
    }
}

#' Anthropic chat request
#' @noRd
.chat_anthropic <- function(body, config, stream, cache = "none",
                            thinking_budget_tokens = NULL) {
    url <- paste0(config$base_url, config$chat_path)

    # Convert messages format for Anthropic
    system_msg <- NULL
    messages <- list()

    for (msg in body$messages) {
        if (msg$role == "system") {
            system_msg <- msg$content
        } else {
            messages[[length(messages) + 1]] <- msg
        }
    }

    anthropic_body <- list(model = body$model, messages = messages,
                           max_tokens = body$max_tokens %||% 4096)

    if (!is.null(system_msg)) {
        anthropic_body$system <- .anthropic_system_with_cache(system_msg, cache)
    }

    if (!is.null(body$temperature)) {
        anthropic_body$temperature <- body$temperature
    }

    if (!is.null(thinking_budget_tokens)) {
        anthropic_body$thinking <- list(
                                        type = "enabled",
                                        budget_tokens = as.integer(thinking_budget_tokens)
        )
    }

    headers <- c(
                 "Content-Type" = "application/json",
                 "x-api-key" = config$api_key,
                 "anthropic-version" = "2023-06-01"
    )

    h <- curl::new_handle()
    curl::handle_setopt(h,
                        customrequest = "POST",
                        postfields = jsonlite::toJSON(anthropic_body, auto_unbox = TRUE)
    )
    curl::handle_setheaders(h, .list = as.list(headers))

    resp <- curl::curl_fetch_memory(url, handle = h)

    if (resp$status_code >= 400) {
        err <- tryCatch(
                        jsonlite::fromJSON(rawToChar(resp$content)),
                        error = function(e) list(error = list(message = rawToChar(resp$content)))
        )
        stop("API error (", resp$status_code, "): ",
             err$error$message %||% "Unknown error", call. = FALSE)
    }

    data <- jsonlite::fromJSON(rawToChar(resp$content))

    # Handle both data.frame and list formats from jsonlite. content is an
    # ordered list of blocks; pull text out of "text" blocks and thinking
    # out of "thinking" blocks.
    if (is.data.frame(data$content)) {
        types <- data$content$type
        text_blocks <- data$content$text[types == "text"]
        thinking_blocks <- data$content$thinking[types == "thinking"]
    } else {
        types <- vapply(data$content, function(b) b$type %||% "", character(1))
        text_blocks <- vapply(data$content[types == "text"],
                              function(b) b$text %||% "", character(1))
        thinking_blocks <- vapply(data$content[types == "thinking"],
                                  function(b) b$thinking %||% "", character(1))
    }

    if (length(text_blocks)) {
        content <- paste(text_blocks, collapse = "\n")
    } else {
        content <- ""
    }
    thinking <- if (length(thinking_blocks)) {
        paste(thinking_blocks, collapse = "\n")
    } else {
        NULL
    }
    finish_reason <- .normalize_anthropic_stop_reason(data$stop_reason)

    .warn_if_truncated(content, thinking, finish_reason)

    list(
         content = content,
         thinking = thinking,
         finish_reason = finish_reason,
         usage = data$usage
    )
}

# Map Anthropic's stop_reason to OpenAI-style finish_reason so callers see
# one vocabulary across providers. "max_tokens" is Anthropic's name for
# what OpenAI calls "length"; "end_turn" maps to "stop". Other values
# ("stop_sequence", "tool_use", "pause_turn", "refusal") pass through.
.normalize_anthropic_stop_reason <- function(stop_reason) {
    if (is.null(stop_reason) || !nzchar(stop_reason)) {
        return(NULL)
    }
    switch(stop_reason, "end_turn" = "stop", "max_tokens" = "length",
           stop_reason)
}

# Surface the silent-empty-content failure mode of reasoning models. When
# the model burns its budget on chain-of-thought without ever emitting a
# user-facing answer, callers otherwise see content="" and assume the
# model decided to say nothing.
.warn_if_truncated <- function(content, thinking, finish_reason) {
    if (identical(finish_reason, "length") &&
        !nzchar(content %||% "") &&
        nzchar(thinking %||% "")) {
        warning("Model truncated mid-reasoning; partial chain-of-thought ",
                "available in $thinking. Increase max_tokens.", call. = FALSE)
    }
}

#' Stream response with live output
#' @noRd
.stream_response <- function(url, handle) {
    full_content <- ""
    full_thinking <- ""
    finish_reason <- NULL

    callback <- function(data) {
        lines <- strsplit(rawToChar(data), "\n")[[1]]
        for (line in lines) {
            if (startsWith(line, "data: ") && line != "data: [DONE]") {
                json_str <- substring(line, 7)
                tryCatch({
                    chunk <- jsonlite::fromJSON(json_str)
                    choice <- chunk$choices[[1]]
                    delta <- choice$delta$content
                    if (!is.null(delta)) {
                        cat(delta)
                        full_content <<- paste0(full_content, delta)
                    }
                    think_delta <- choice$delta$reasoning_content %||%
                    choice$delta$reasoning
                    if (!is.null(think_delta)) {
                        full_thinking <<- paste0(full_thinking, think_delta)
                    }
                    if (!is.null(choice$finish_reason)) {
                        finish_reason <<- choice$finish_reason
                    }
                }, error = function(e) NULL)
            }
        }
        length(data)
    }

    curl::handle_setopt(handle, writefunction = callback)
    curl::curl_fetch_memory(url, handle = handle)
    cat("\n")

    if (nzchar(full_thinking)) {
        thinking <- full_thinking
    } else {
        thinking <- NULL
    }
    .warn_if_truncated(full_content, thinking, finish_reason)

    list(content = full_content, thinking = thinking,
         finish_reason = finish_reason, usage = NULL)
}

#' Null coalescing operator
#' @noRd
`%||%` <- function(x, y) if (is.null(x)) {
    y
} else {
    x
}

# Tests for the 0.1.4 additions:
#  - Anthropic cache_control on system message
#  - Anthropic thinking_budget_tokens (validation + body shape)
#  - OpenAI max_tokens -> max_completion_tokens rename (NOT Moonshot)
#
# `agent()` uses .post_json which we can stub cleanly; `chat()` uses
# curl::curl_fetch_memory directly so we test those code paths via
# the extracted body-builder helpers (.anthropic_system_with_cache,
# .validate_thinking_budget) and lean on agent() for integration
# coverage of the same logic.

ns <- asNamespace("llm.api")
orig_post_json <- get(".post_json", envir = ns, inherits = FALSE)

with_stubbed_post_json <- function(stub, expr) {
    assignInNamespace(".post_json", stub, ns = "llm.api")
    tryCatch(force(expr),
             finally = assignInNamespace(".post_json", orig_post_json,
                                         ns = "llm.api"))
}

# --- .anthropic_system_with_cache helper ----------------------------

# cache = "none" returns the system text unchanged.
expect_equal(llm.api:::.anthropic_system_with_cache("be helpful", "none"),
             "be helpful")

# cache = "5m" wraps in a single text block with ephemeral cache_control
# and no ttl (5min is Anthropic's default ephemeral TTL).
local({
    out <- llm.api:::.anthropic_system_with_cache("be helpful", "5m")
    expect_true(is.list(out))
    expect_equal(length(out), 1L)
    expect_equal(out[[1L]]$type, "text")
    expect_equal(out[[1L]]$text, "be helpful")
    expect_equal(out[[1L]]$cache_control$type, "ephemeral")
    expect_null(out[[1L]]$cache_control$ttl)
})

# cache = "1h" sets ttl = "1h" on the cache_control block.
local({
    out <- llm.api:::.anthropic_system_with_cache("be helpful", "1h")
    expect_equal(out[[1L]]$cache_control$type, "ephemeral")
    expect_equal(out[[1L]]$cache_control$ttl, "1h")
})

# --- .validate_thinking_budget --------------------------------------

# Must be a positive integer.
expect_error(llm.api:::.validate_thinking_budget("not a number"),
             pattern = "must be a single integer")
expect_error(llm.api:::.validate_thinking_budget(c(1024L, 2048L)),
             pattern = "must be a single integer")
expect_error(llm.api:::.validate_thinking_budget(NA),
             pattern = "must be a single integer")
expect_error(llm.api:::.validate_thinking_budget(1500.5),
             pattern = "must be a single integer")

# Must be at least 1024 (Anthropic's minimum).
expect_error(llm.api:::.validate_thinking_budget(512L),
             pattern = "at least 1024")

# Budget must leave room for the completion.
expect_error(
             llm.api:::.validate_thinking_budget(8000L, max_tokens = 8000L),
             pattern = "strictly less than .max_tokens."
)
expect_error(
             llm.api:::.validate_thinking_budget(8000L, max_tokens = 4000L),
             pattern = "strictly less than .max_tokens."
)

# Valid input -> invisible TRUE.
expect_true(llm.api:::.validate_thinking_budget(2048L, max_tokens = 8000L))
expect_true(llm.api:::.validate_thinking_budget(2048L, max_tokens = NULL))

# --- chat() rejects Anthropic-only features on other providers ------

# chat(provider = "openai", cache = "5m") warns and downgrades to none.
# We don't need to make the request succeed; we just want to catch
# the warning.
local({
    expect_warning(
        out <- tryCatch(
                        llm.api::chat(prompt = "x", provider = "openai",
                                      cache = "5m"),
                        error = function(e) NULL
        ),
        pattern = "cache.*Anthropic-only"
    )
})

# chat(provider = "openai", thinking_budget_tokens = 2048) warns and
# drops the param.
local({
    expect_warning(
        out <- tryCatch(
                        llm.api::chat(prompt = "x", provider = "openai",
                                      max_tokens = 8000L,
                                      thinking_budget_tokens = 2048L),
                        error = function(e) NULL
        ),
        pattern = "thinking_budget_tokens.*Anthropic-only"
    )
})

# --- agent() Anthropic with cache + thinking_budget_tokens ----------

# agent() integration: cache and thinking_budget_tokens make it into
# the body Anthropic sees. Uses the .post_json stub.
anth_agent_capture <- NULL
anth_agent_stub <- function(url, body, headers) {
    anth_agent_capture <<- body
    list(
         content = list(list(type = "text", text = "done")),
         usage = list(input_tokens = 1L, output_tokens = 1L)
    )
}

local({
    anth_agent_capture <<- NULL
    with_stubbed_post_json(anth_agent_stub, {
        llm.api::agent(prompt = "go",
                       provider = "anthropic",
                       system = "system text",
                       model = "claude-test",
                       cache = "5m",
                       thinking_budget_tokens = 2048L,
                       verbose = FALSE,
                       tools = list())
    })
    expect_true(is.list(anth_agent_capture$system))
    expect_equal(anth_agent_capture$system[[1L]]$cache_control$type,
                 "ephemeral")
    expect_equal(anth_agent_capture$thinking$type, "enabled")
    expect_equal(anth_agent_capture$thinking$budget_tokens, 2048L)
})

# agent() with cache = "1h" sets the longer TTL.
local({
    anth_agent_capture <<- NULL
    with_stubbed_post_json(anth_agent_stub, {
        llm.api::agent(prompt = "go",
                       provider = "anthropic",
                       system = "system text",
                       model = "claude-test",
                       cache = "1h",
                       verbose = FALSE,
                       tools = list())
    })
    expect_equal(anth_agent_capture$system[[1L]]$cache_control$ttl, "1h")
})

# agent() default: no cache, no thinking -- system stays as plain text.
local({
    anth_agent_capture <<- NULL
    with_stubbed_post_json(anth_agent_stub, {
        llm.api::agent(prompt = "go",
                       provider = "anthropic",
                       system = "system text",
                       model = "claude-test",
                       verbose = FALSE,
                       tools = list())
    })
    expect_equal(anth_agent_capture$system, "system text")
    expect_null(anth_agent_capture$thinking)
})

# --- agent() OpenAI max_tokens rename via ... ----------------------

# Capture the body that .agent_openai sends; verify max_tokens
# becomes max_completion_tokens for openai and stays as max_tokens
# for moonshot.
openai_capture_body <- NULL
openai_capture_stub <- function(url, body, headers) {
    openai_capture_body <<- body
    list(
         choices = list(list(
                             message = list(role = "assistant", content = "done",
                                            tool_calls = NULL),
                             finish_reason = "stop"
        )),
         usage = list(prompt_tokens = 1L, completion_tokens = 1L)
    )
}

# OpenAI: max_tokens via ... gets renamed to max_completion_tokens.
local({
    openai_capture_body <<- NULL
    with_stubbed_post_json(openai_capture_stub, {
        llm.api::agent(prompt = "go",
                       provider = "openai",
                       model = "gpt-4o",
                       verbose = FALSE,
                       max_tokens = 200L)
    })
    expect_null(openai_capture_body$max_tokens)
    expect_equal(openai_capture_body$max_completion_tokens, 200L)
})

# Moonshot: max_tokens via ... stays as max_tokens.
local({
    openai_capture_body <<- NULL
    with_stubbed_post_json(openai_capture_stub, {
        llm.api::agent(prompt = "go",
                       provider = "moonshot",
                       model = "kimi-k2",
                       verbose = FALSE,
                       max_tokens = 200L)
    })
    expect_equal(openai_capture_body$max_tokens, 200L)
    expect_null(openai_capture_body$max_completion_tokens)
})

# OpenAI: caller-provided max_completion_tokens takes precedence;
# max_tokens is left for the caller to disambiguate. The rename is
# gated on is.null(b$max_completion_tokens).
local({
    openai_capture_body <<- NULL
    with_stubbed_post_json(openai_capture_stub, {
        llm.api::agent(prompt = "go",
                       provider = "openai",
                       model = "gpt-4o",
                       verbose = FALSE,
                       max_completion_tokens = 500L,
                       max_tokens = 200L)
    })
    expect_equal(openai_capture_body$max_completion_tokens, 500L)
})

# --- chat() resolves provider BEFORE the Anthropic-only guards -------
# Regression (0.1.3.3): the cache / thinking_budget_tokens guards used
# to run against the literal "auto" default, before .detect_provider(),
# so a Claude model under the default provider had its opt-ins silently
# downgraded with a spurious "Anthropic-only" warning. Detection now
# runs first. We stub chat()'s network dispatch so only the guard logic
# is exercised -- no key, no network, runs under R CMD check.

collect_warnings <- function(expr) {
    w <- character(0)
    withCallingHandlers(
        force(expr),
        warning = function(cond) {
            w[[length(w) + 1L]] <<- conditionMessage(cond)
            invokeRestart("muffleWarning")
        })
    w
}

with_stubbed <- function(name, stub, expr) {
    orig <- get(name, envir = ns, inherits = FALSE)
    assignInNamespace(name, stub, ns = "llm.api")
    tryCatch(force(expr),
             finally = assignInNamespace(name, orig, ns = "llm.api"))
}

captured <- NULL
anthropic_stub <- function(body, config, stream, cache = "none",
                           thinking_budget_tokens = NULL) {
    captured <<- list(cache = cache, thinking = thinking_budget_tokens)
    list(content = "ok", thinking = NULL, finish_reason = "stop", usage = NULL)
}

# Claude model + default ("auto") provider + cache: no warning, and the
# cache value reaches the dispatch instead of being zeroed to "none".
local({
    captured <<- NULL
    w <- collect_warnings(with_stubbed(".chat_anthropic", anthropic_stub, {
        llm.api::chat("hi", model = "claude-sonnet-4-6", cache = "5m")
    }))
    expect_false(any(grepl("Anthropic-only", w)))
    expect_equal(captured$cache, "5m")
})

# Same for thinking_budget_tokens.
local({
    captured <<- NULL
    w <- collect_warnings(with_stubbed(".chat_anthropic", anthropic_stub, {
        llm.api::chat("hi", model = "claude-sonnet-4-6",
                      thinking_budget_tokens = 2000L, max_tokens = 4000L)
    }))
    expect_false(any(grepl("Anthropic-only", w)))
    expect_equal(captured$thinking, 2000L)
})

# Positive control: a genuine OpenAI model still trips the guard and the
# warning fires (detection resolves "auto" -> openai before the guard).
local({
    openai_stub <- function(body, config, stream) {
        list(content = "ok", thinking = NULL, finish_reason = "stop",
             usage = NULL)
    }
    w <- collect_warnings(with_stubbed(".chat_openai_compatible", openai_stub, {
        llm.api::chat("hi", model = "gpt-4o", cache = "5m")
    }))
    expect_true(any(grepl("cache.*Anthropic-only", w)))
})

# Tests for reasoning_content / thinking handling.

# Anthropic stop_reason → finish_reason normalization. Anthropic uses
# different vocabulary than OpenAI; chat() exposes the OpenAI vocabulary
# uniformly, so callers don't have to branch on provider.
expect_null(llm.api:::.normalize_anthropic_stop_reason(NULL))
expect_null(llm.api:::.normalize_anthropic_stop_reason(""))
expect_equal(llm.api:::.normalize_anthropic_stop_reason("end_turn"), "stop")
expect_equal(llm.api:::.normalize_anthropic_stop_reason("max_tokens"), "length")
# Pass-through for values that don't have an OpenAI equivalent.
expect_equal(llm.api:::.normalize_anthropic_stop_reason("tool_use"), "tool_use")
expect_equal(llm.api:::.normalize_anthropic_stop_reason("stop_sequence"),
             "stop_sequence")

# Truncation warning fires only when content is empty AND thinking is
# populated AND finish_reason is "length". This is the silent-failure
# mode of reasoning models that callers need to know about.
expect_warning(
    llm.api:::.warn_if_truncated("", "some reasoning", "length"),
    pattern = "truncated mid-reasoning"
)
# Normal completion with both content and thinking: no warning.
expect_silent(
    llm.api:::.warn_if_truncated("answer", "thinking", "stop")
)
# Length-truncated but content is non-empty: model hit the cap mid-answer
# but at least said something. Caller can decide; we don't warn here.
expect_silent(
    llm.api:::.warn_if_truncated("partial answer", "thinking", "length")
)
# Length-truncated with no thinking and no content: not a reasoning model
# truncation pattern, no warning.
expect_silent(
    llm.api:::.warn_if_truncated("", NULL, "length")
)
# Length-truncated with thinking but content is NULL (some providers).
expect_warning(
    llm.api:::.warn_if_truncated(NULL, "some reasoning", "length"),
    pattern = "truncated mid-reasoning"
)

# MCP (Model Context Protocol) client
#
# Connects to MCP servers via socket transport (base R)

#' Connect to an MCP server
#'
#' Connects to an MCP server via TCP socket.
#'
#' @param host Character. Server hostname (default: "localhost").
#' @param port Integer. Server port.
#' @param name Character. Friendly name for this server.
#' @param timeout Numeric. Connection timeout in seconds (default: 30).
#'
#' @return An MCP connection object (list with socket and tools).
#' @export
#'
#' @examples
#' \dontrun{
#' # Start server first: r mcp_server.R --port 7850
#' conn <- mcp_connect(port = 7850, name = "codeR")
#' tools <- mcp_tools(conn)
#' result <- mcp_call(conn, "read_file", list(path = "README.md"))
#' mcp_close(conn)
#' }
mcp_connect <- function(host = "localhost", port, name = NULL, timeout = 30) {
    # Connect via socket
    sock <- tryCatch(
                     socketConnection(host = host, port = port, blocking = TRUE,
                                      open = "r+b", timeout = timeout),
                     error = function(e) {
        stop("Failed to connect to MCP server at ", host, ":", port,
             "\n  ", e$message, call. = FALSE)
    }
    )

    conn <- list(
                 socket = sock,
                 host = host,
                 port = port,
                 name = name %||% paste0(host, ":", port),
                 tools = list(),
                 request_id = 0L
    )
    class(conn) <- "mcp_connection"

    # Initialize handshake
    init_result <- .mcp_request(conn, "initialize", list(
            protocolVersion = "2024-11-05",
            capabilities = list(),
            clientInfo = list(name = "llm.api", version = "0.1.0")
        ))

    # Send initialized notification
    .mcp_notify(conn, "notifications/initialized", list())

    # Get tools
    tools_result <- .mcp_request(conn, "tools/list", list())
    conn$tools <- tools_result$tools

    conn
}

#' Start and connect to an MCP server
#'
#' Spawns an MCP server process and connects to it.
#' Requires the server script to support --port argument.
#'
#' @param command Character. Command to run the server (e.g., "r", "Rscript").
#' @param args Character vector. Arguments (path to server script).
#' @param port Integer. Port for the server (default: random 7850-7899).
#' @param name Character. Friendly name.
#' @param startup_wait Numeric. Seconds to wait for server startup.
#'
#' @return An MCP connection object.
#' @export
#' @examples
#' \dontrun{
#' conn <- mcp_start("Rscript", args = "mcp_server.R", port = 7850)
#' mcp_close(conn)
#' }
mcp_start <- function(command, args = character(), port = NULL, name = NULL,
                      startup_wait = 2) {
    # Pick random port if not specified
    if (is.null(port)) {
        port <- sample(7850:7899, 1)
    }

    # Add port argument
    full_args <- c(args, "--port", as.character(port))

    # Start server in background
    if (.Platform$OS.type == "windows") {
        system2(command, full_args, wait = FALSE, stdout = FALSE,
                stderr = FALSE)
    } else {
        system2(command, full_args, wait = FALSE,
                stdout = "/dev/null", stderr = "/dev/null")
    }

    # Wait for server to start
    Sys.sleep(startup_wait)

    # Connect
    mcp_connect(port = port, name = name %||% basename(args[1]))
}

#' List tools from an MCP connection
#'
#' @param conn An MCP connection object.
#' @return List of tool definitions.
#' @export
#' @examples
#' \dontrun{
#' conn <- mcp_connect(port = 7850)
#' tools <- mcp_tools(conn)
#' mcp_close(conn)
#' }
mcp_tools <- function(conn) {
    conn$tools
}

#' Call a tool on an MCP server
#'
#' @param conn An MCP connection object.
#' @param name Character. Tool name.
#' @param arguments List. Tool arguments.
#'
#' @return Tool result (list with content and text).
#' @export
#' @examples
#' \dontrun{
#' conn <- mcp_connect(port = 7850)
#' result <- mcp_call(conn, "read_file", list(path = "README.md"))
#' mcp_close(conn)
#' }
mcp_call <- function(conn, name, arguments = list()) {
    result <- .mcp_request(conn, "tools/call",
                           list(name = name, arguments = arguments))

    # Extract text content
    if (!is.null(result$content)) {
        texts <- vapply(result$content, function(c) {
            if (identical(c$type, "text")) c$text else ""
        }, character(1))
        result$text <- paste(texts, collapse = "\n")
    }

    result
}

#' Close an MCP connection
#'
#' @param conn An MCP connection object.
#' @return \code{NULL}, invisibly. Called for its side effect of closing
#'   the underlying socket.
#' @export
#' @examples
#' \dontrun{
#' conn <- mcp_connect(port = 7850)
#' mcp_close(conn)
#' }
mcp_close <- function(conn) {
    tryCatch(close(conn$socket), error = function(e) NULL)
    invisible(NULL)
}

#' Format MCP tools for LLM APIs
#'
#' Converts MCP tool definitions to the format used by Claude/OpenAI.
#'
#' @param conn An MCP connection, or list of connections.
#' @return List of tools in API format.
#' @export
#' @examples
#' \dontrun{
#' conn <- mcp_connect(port = 7850)
#' tools <- mcp_tools_for_api(conn)
#' mcp_close(conn)
#' }
mcp_tools_for_api <- function(conn) {
    if (inherits(conn, "mcp_connection")) {
        conns <- list(conn)
    } else {
        conns <- conn
    }

    tools <- list()
    for (c in conns) {
        for (tool in c$tools) {
            tools[[length(tools) + 1]] <- list(name = tool$name,
                description = tool$description %||% "",
                input_schema = tool$inputSchema)
        }
    }
    tools
}

#' Format MCP tools for Claude API
#'
#' Wrapper for \code{\link{mcp_tools_for_api}}, retained for backwards
#' compatibility.
#'
#' @param conn An MCP connection, or list of connections.
#' @return List of tools in API format.
#' @export
#' @examples
#' \dontrun{
#' conn <- mcp_connect(host = "localhost", port = 7850)
#' tools <- mcp_tools_for_claude(conn)
#' mcp_close(conn)
#' }
mcp_tools_for_claude <- function(conn) {
    mcp_tools_for_api(conn)
}

# Internal: send JSON-RPC request and get response
.mcp_request <- function(conn, method, params) {
    conn$request_id <- conn$request_id + 1L
    id <- conn$request_id

    request <- list(jsonrpc = "2.0", id = id, method = method, params = params)

    json <- jsonlite::toJSON(request, auto_unbox = TRUE, null = "null")

    # Write request
    writeLines(json, conn$socket)

    # Read response
    response_line <- readLines(conn$socket, n = 1, warn = FALSE)

    if (length(response_line) == 0 || nchar(response_line) == 0) {
        stop("MCP server closed connection", call. = FALSE)
    }

    response <- jsonlite::fromJSON(response_line, simplifyVector = FALSE)

    if (!is.null(response$error)) {
        stop("MCP error: ", response$error$message, call. = FALSE)
    }

    response$result
}

# Internal: send notification (no response expected)
.mcp_notify <- function(conn, method, params) {
    request <- list(jsonrpc = "2.0", method = method, params = params)

    json <- jsonlite::toJSON(request, auto_unbox = TRUE, null = "null")
    writeLines(json, conn$socket)
    invisible(NULL)
}

#' Print an MCP Connection
#'
#' S3 print method for MCP connection objects.
#'
#' @param x An MCP connection object.
#' @param ... Unused.
#' @return \code{x}, invisibly. Called for the side effect of printing a
#'   summary of the connection state and available tools.
#' @export
#' @examples
#' \dontrun{
#' conn <- mcp_connect(port = 7850)
#' print(conn)
#' mcp_close(conn)
#' }
print.mcp_connection <- function(x, ...) {
    status <- tryCatch({
        if (isOpen(x$socket)) "connected" else "disconnected"
    }, error = function(e) "disconnected")

    cat(sprintf("MCP Connection: %s (%s)\n", x$name, status))
    cat(sprintf("  Endpoint: %s:%d\n", x$host, x$port))
    cat(sprintf("  Tools: %d\n", length(x$tools)))
    if (length(x$tools) > 0) {
        for (tool in x$tools) {
            cat(sprintf("    - %s\n", tool$name))
        }
    }
    invisible(x)
}

# Helpers for walking the message history that agent() returns.
#
# `history` is provider-native by design: it's the sequence of messages
# we send back into the next API call, so the shape has to match what
# each provider expects on input. That means consumers walking history
# for tool-call introspection have to handle two distinct shapes:
#
#   Anthropic
#     assistant role, content = list of blocks, with
#       {type = "tool_use", id, name, input}
#     user role, content = list of blocks, with
#       {type = "tool_result", tool_use_id, content}
#
#   OpenAI / moonshot / ollama
#     assistant role, content = "" (or text), tool_calls = list of
#       {id, type = "function", function = {name, arguments}}
#     tool role, tool_call_id, name, content
#
# These helpers pair calls with their results and return a single
# canonical record list so consumers stop reinventing the walk.

#' Walk a history list and return paired tool-call / tool-result records.
#'
#' Accepts either a `history` list as returned by [agent()] (in which
#' case every entry is treated as a message) or any list-of-messages
#' that follows the same shape. Returns a list of records, one per
#' tool call, each with:
#'
#' \describe{
#'   \item{id}{Canonical call id (synthesized for Ollama responses
#'     that omit one).}
#'   \item{name}{Tool name.}
#'   \item{arguments}{Argument list. Parsed from JSON for OpenAI-style
#'     shapes; passed through for Anthropic.}
#'   \item{result}{Tool result text, or `NULL` if the call has no
#'     matching result yet.}
#'   \item{completed}{`TRUE` when `result` is non-`NULL`.}
#'   \item{call_message_index}{1-based index of the assistant message
#'     that issued the call.}
#'   \item{result_message_index}{1-based index of the message
#'     carrying the result, or `NA_integer_` for unfinished calls.}
#'   \item{provider_shape}{`"anthropic"` or `"openai"`. Useful when a
#'     consumer needs to branch on shape (rare).}
#' }
#'
#' @param history List of messages, typically the `history` element
#'   from an [agent()] return value.
#' @return A list of tool-call records (possibly empty). Records are
#'   returned in the order calls were issued.
#' @export
history_tool_calls <- function(history) {
    if (!is.list(history) || length(history) == 0L) {
        return(list())
    }

    # Pass 1: collect calls.
    calls <- list()
    for (i in seq_along(history)) {
        entry <- history[[i]]
        if (!is.list(entry)) {
            next
        }
        role <- entry$role %||% ""
        if (!identical(role, "assistant")) {
            next
        }

        cnt <- entry$content
        # Anthropic shape: content is a list of typed blocks.
        if (is.list(cnt)) {
            for (block in cnt) {
                if (identical(block$type %||% "", "tool_use")) {
                    calls[[length(calls) + 1L]] <- list(id = block$id %||% "",
                        name = block$name %||% "",
                        arguments = block$input %||% list(), result = NULL,
                        completed = FALSE, call_message_index = i,
                        result_message_index = NA_integer_,
                        provider_shape = "anthropic")
                }
            }
        }
        # OpenAI shape: separate tool_calls field on the assistant message.
        if (!is.null(entry$tool_calls)) {
            for (tc in entry$tool_calls) {
                fn <- tc$`function` %||% list()
                args_raw <- fn$arguments %||% list()
                args <- if (is.character(args_raw) && length(args_raw) == 1L) {
                    tryCatch(
                             jsonlite::fromJSON(args_raw, simplifyVector = FALSE),
                             error = function(e) list()
                    )
                } else {
                    args_raw
                }
                calls[[length(calls) + 1L]] <- list(
                    id = tc$id %||% "",
                    name = fn$name %||% "",
                    arguments = args,
                    result = NULL,
                    completed = FALSE,
                    call_message_index = i,
                    result_message_index = NA_integer_,
                    provider_shape = "openai"
                )
            }
        }
    }

    if (length(calls) == 0L) {
        return(list())
    }

    # Pass 2: pair results back to calls.
    for (i in seq_along(history)) {
        entry <- history[[i]]
        if (!is.list(entry)) {
            next
        }
        role <- entry$role %||% ""

        # Anthropic: tool_result blocks live in user-role messages.
        if (identical(role, "user")) {
            cnt <- entry$content
            if (is.list(cnt)) {
                for (block in cnt) {
                    if (identical(block$type %||% "", "tool_result")) {
                        target_id <- block$tool_use_id %||% block$id %||% ""
                        if (!nzchar(target_id)) {
                            next
                        }
                        result_text <- .history_block_result_text(block)
                        for (j in seq_along(calls)) {
                            if (!calls[[j]]$completed &&
                                identical(calls[[j]]$id, target_id)) {
                                calls[[j]]$result <- result_text
                                calls[[j]]$completed <- TRUE
                                calls[[j]]$result_message_index <- i
                                break
                            }
                        }
                    }
                }
            }
            next
        }

        # OpenAI: each tool result is its own role="tool" message.
        if (identical(role, "tool")) {
            target_id <- entry$tool_call_id %||% entry$id %||% ""
            if (!nzchar(target_id)) {
                next
            }
            result_text <- as.character(entry$content %||% "")
            for (j in seq_along(calls)) {
                if (!calls[[j]]$completed &&
                    identical(calls[[j]]$id, target_id)) {
                    calls[[j]]$result <- result_text
                    calls[[j]]$completed <- TRUE
                    calls[[j]]$result_message_index <- i
                    break
                }
            }
        }
    }

    calls
}

#' Count tool calls in a history list.
#'
#' Thin convenience wrapper over [history_tool_calls()]. Pass
#' `completed_only = TRUE` to count only calls that have matching
#' results (i.e., to skip mid-flight calls at the end of a turn that
#' got cut off).
#'
#' @param history List of messages.
#' @param completed_only Logical. When `TRUE`, count only calls whose
#'   result is present in the history. Default `FALSE`.
#' @return Single integer.
#' @export
history_count_tool_calls <- function(history, completed_only = FALSE) {
    calls <- history_tool_calls(history)
    if (length(calls) == 0L) {
        return(0L)
    }
    if (isTRUE(completed_only)) {
        completed <- vapply(calls, function(c) isTRUE(c$completed), logical(1))
        return(as.integer(sum(completed)))
    }
    length(calls)
}

# ---- Internal helpers ----

# Anthropic tool_result blocks may be a flat string or a list of inner
# {type: "text"} blocks. Render to a single string.
.history_block_result_text <- function(block) {
    cnt <- block$content
    if (is.character(cnt)) {
        return(paste(cnt, collapse = "\n"))
    }
    if (is.list(cnt)) {
        parts <- vapply(cnt, function(b) {
            as.character(b$text %||% "")
        }, character(1))
        return(paste(parts, collapse = "\n"))
    }
    as.character(cnt %||% "")
}

# Note: `%||%` is defined in chat.R for the package; reused here.

# Tests for history_tool_calls + history_count_tool_calls.
# Pure-function, offline. Exercises both Anthropic and OpenAI shapes
# plus the unfinished-call edge case.

# ---- Empty / malformed inputs ----

expect_equal(llm.api::history_tool_calls(list()), list())
expect_equal(llm.api::history_tool_calls(NULL), list())
expect_equal(llm.api::history_count_tool_calls(list()), 0L)

# ---- Anthropic shape ----

anthropic_history <- list(
    list(role = "user", content = "find auth"),
    list(role = "assistant", content = list(
        list(type = "text", text = "Searching."),
        list(type = "tool_use", id = "tu1", name = "grep_files",
             input = list(pattern = "auth"))
    )),
    list(role = "user", content = list(
        list(type = "tool_result", tool_use_id = "tu1",
             content = list(list(type = "text", text = "src/auth.R")))
    )),
    list(role = "assistant", content = list(
        list(type = "tool_use", id = "tu2", name = "read_file",
             input = list(path = "src/auth.R"))
    )),
    list(role = "user", content = list(
        list(type = "tool_result", tool_use_id = "tu2",
             content = "auth_handler <- function() {}")
    ))
)

calls <- llm.api::history_tool_calls(anthropic_history)
expect_equal(length(calls), 2L)

expect_equal(calls[[1]]$id, "tu1")
expect_equal(calls[[1]]$name, "grep_files")
expect_equal(calls[[1]]$arguments$pattern, "auth")
expect_equal(calls[[1]]$result, "src/auth.R")
expect_true(calls[[1]]$completed)
expect_equal(calls[[1]]$call_message_index, 2L)
expect_equal(calls[[1]]$result_message_index, 3L)
expect_equal(calls[[1]]$provider_shape, "anthropic")

expect_equal(calls[[2]]$id, "tu2")
expect_equal(calls[[2]]$result, "auth_handler <- function() {}")

expect_equal(llm.api::history_count_tool_calls(anthropic_history), 2L)

# ---- OpenAI shape (covers moonshot / kimi / ollama) ----

openai_history <- list(
    list(role = "user", content = "find auth"),
    list(role = "assistant", content = "",
         tool_calls = list(
             list(id = "c1", type = "function",
                  `function` = list(name = "grep_files",
                                    arguments = "{\"pattern\":\"auth\"}")),
             list(id = "c2", type = "function",
                  `function` = list(name = "read_file",
                                    arguments = "{\"path\":\"src/auth.R\"}"))
         )),
    list(role = "tool", tool_call_id = "c1",
         name = "grep_files", content = "src/auth.R"),
    list(role = "tool", tool_call_id = "c2",
         name = "read_file", content = "auth_handler <- function() {}")
)

calls <- llm.api::history_tool_calls(openai_history)
expect_equal(length(calls), 2L)

expect_equal(calls[[1]]$id, "c1")
expect_equal(calls[[1]]$name, "grep_files")
# arguments parsed from JSON
expect_equal(calls[[1]]$arguments$pattern, "auth")
expect_equal(calls[[1]]$result, "src/auth.R")
expect_true(calls[[1]]$completed)
expect_equal(calls[[1]]$provider_shape, "openai")

expect_equal(calls[[2]]$arguments$path, "src/auth.R")
expect_equal(calls[[2]]$result, "auth_handler <- function() {}")

expect_equal(llm.api::history_count_tool_calls(openai_history), 2L)

# ---- Unfinished call (assistant emitted tool_use, no matching result) ----

unfinished_anthropic <- list(
    list(role = "user", content = "do thing"),
    list(role = "assistant", content = list(
        list(type = "tool_use", id = "tu_x", name = "bash",
             input = list(command = "ls"))
    ))
)
calls <- llm.api::history_tool_calls(unfinished_anthropic)
expect_equal(length(calls), 1L)
expect_false(calls[[1]]$completed)
expect_null(calls[[1]]$result)
expect_true(is.na(calls[[1]]$result_message_index))

# completed_only filters out the unfinished one
expect_equal(llm.api::history_count_tool_calls(unfinished_anthropic), 1L)
expect_equal(
    llm.api::history_count_tool_calls(unfinished_anthropic,
                                       completed_only = TRUE),
    0L
)

unfinished_openai <- list(
    list(role = "user", content = "do thing"),
    list(role = "assistant", content = "",
         tool_calls = list(
             list(id = "c_x", type = "function",
                  `function` = list(name = "bash",
                                    arguments = "{\"command\":\"ls\"}"))
         ))
)
calls <- llm.api::history_tool_calls(unfinished_openai)
expect_equal(length(calls), 1L)
expect_false(calls[[1]]$completed)

# ---- Mixed shapes in a single history would never happen in practice
# (a single agent call sticks to one provider), but the helper should
# at least not crash on them.

mixed <- c(anthropic_history, openai_history)
calls <- llm.api::history_tool_calls(mixed)
expect_equal(length(calls), 4L)

# ---- A user message that's a plain string contributes nothing ----

trivial <- list(
    list(role = "user", content = "hello"),
    list(role = "assistant", content = "hi back")
)
expect_equal(length(llm.api::history_tool_calls(trivial)), 0L)
expect_equal(llm.api::history_count_tool_calls(trivial), 0L)

# Stateful chat session wrapper
# Provides ellmer-compatible interface for use with shinychat

#' Create a stateful chat session
#'
#' Creates a chat session that maintains conversation history internally.
#' Returns a list of functions for interacting with the session.
#'
#' @param model Character. Model name.
#' @param system_prompt Character or NULL. System prompt.
#' @param provider Character. Provider: "openai", "anthropic", "moonshot",
#'   or "ollama".
#' @param ... Additional parameters passed to chat().
#'
#' @return A list with functions:
#'   \item{chat(message)}{Send a message, returns response text}
#'   \item{stream(message)}{Stream a response, returns response text}
#'   \item{stream_async(message)}{Async stream for shinychat (returns string)}
#'   \item{last_turn()}{Get the last response as list(role, text)}
#'   \item{history()}{Get full conversation history}
#'   \item{clear()}{Clear conversation history}
#'
#' @export
#' @examples
#' \dontrun{
#' # Create a session
#' session <- chat_session(model = "gpt-5.4-mini", system_prompt = "You are helpful.")
#'
#' # Chat (history maintained automatically)
#' response <- session$chat("Hello")
#' response2 <- session$chat("Tell me more")
#'
#' # Get last response
#' last <- session$last_turn()
#' last$text
#'
#' # Clear and start over
#' session$clear()
#' }
chat_session <- function(model = NULL, system_prompt = NULL,
                         provider = c("openai", "anthropic", "moonshot", "ollama"),
                         ...) {
    provider <- match.arg(provider)

    # Internal state
    .history <- list()
    .system <- system_prompt
    .model <- model
    .provider <- provider
    .extra <- list(...)
    .last_response <- NULL

    # Chat method - send message, return response text
    chat_fn <- function(message) {
        result <- do.call(chat, c(
                                  list(prompt = message, model = .model,
                                       system = if (length(.history) == 0) .system else NULL,
                                       history = .history, provider = .provider, stream = FALSE),
                                  .extra
            ))

        .history <<- result$history
        .last_response <<- list(role = "assistant", text = result$content)

        result$content
    }

    # Stream method - stream response, return full text
    stream_fn <- function(message) {
        result <- do.call(chat, c(
                                  list(
                                       prompt = message,
                                       model = .model,
                                       system = if (length(.history) == 0) .system else NULL,
                                       history = .history,
                                       provider = .provider,
                                       stream = TRUE
                ),
                                  .extra
            ))

        .history <<- result$history
        .last_response <<- list(role = "assistant", text = result$content)

        result$content
    }

    # Async stream for shinychat compatibility
    # Returns string since true async requires coro package
    stream_async_fn <- function(message) {
        chat_fn(message)
    }

    # Get last turn
    last_turn_fn <- function() {
        .last_response
    }

    # Get history
    history_fn <- function() {
        .history
    }

    # Clear history
    clear_fn <- function() {
        .history <<- list()
        .last_response <<- NULL
        invisible(NULL)
    }

    list(
         chat = chat_fn,
         stream = stream_fn,
         stream_async = stream_async_fn,
         last_turn = last_turn_fn,
         history = history_fn,
         clear = clear_fn
    )
}

#' Create OpenAI chat session
#'
#' Convenience wrapper for chat_session with OpenAI provider.
#'
#' @param model Character. Model name (default: "gpt-5.4-mini").
#' @param system_prompt Character or NULL. System prompt.
#' @param ... Additional parameters passed to chat_session().
#'
#' @return A chat session list.
#' @export
#' @examples
#' # Construct a session (no network call yet)
#' session <- chat_session_openai(system_prompt = "You are helpful.")
#' \dontrun{
#' session$chat("Hello")
#' }
chat_session_openai <- function(model = "gpt-5.4-mini", system_prompt = NULL,
                                ...) {
    chat_session(model = model, system_prompt = system_prompt,
                 provider = "openai", ...)
}

#' Create Anthropic chat session
#'
#' Convenience wrapper for chat_session with Anthropic provider.
#'
#' @param model Character. Model name (default: "claude-sonnet-4-6").
#' @param system_prompt Character or NULL. System prompt.
#' @param ... Additional parameters passed to chat_session().
#'
#' @return A chat session list.
#' @export
#' @examples
#' # Construct a session (no network call yet)
#' session <- chat_session_anthropic(system_prompt = "You are helpful.")
#' \dontrun{
#' session$chat("Hello")
#' }
chat_session_anthropic <- function(model = "claude-sonnet-4-6",
                                   system_prompt = NULL, ...) {
    chat_session(model = model, system_prompt = system_prompt,
                 provider = "anthropic", ...)
}

#' Create Ollama chat session
#'
#' Convenience wrapper for chat_session with Ollama provider.
#'
#' @param model Character. Model name (default: "qwen3.5:9b").
#' @param system_prompt Character or NULL. System prompt.
#' @param ... Additional parameters passed to chat_session().
#'
#' @return A chat session list.
#' @export
#' @examples
#' # Construct a session (no network call yet)
#' session <- chat_session_ollama(system_prompt = "You are helpful.")
#' \dontrun{
#' session$chat("Hello")
#' }
chat_session_ollama <- function(model = "qwen3.5:9b", system_prompt = NULL, ...) {
    chat_session(model = model, system_prompt = system_prompt,
                 provider = "ollama", ...)
}

# Anthropic provider-native web search: offline tests for the tool builder and
# the citation/search block parser (the live call needs a key + network).
ns <- asNamespace("llm.api")
`%||%` <- function(x, y) if (is.null(x)) y else x

expect_true("anthropic" %in% ns$.web_search_providers())

# tool builder: off -> NULL; on -> basic web_search_20250305; options mapped
expect_null(ns$.anthropic_web_search_tool(FALSE))
expect_null(ns$.anthropic_web_search_tool(NULL))
t <- ns$.anthropic_web_search_tool(TRUE)
expect_equal(t$type, "web_search_20250305")
expect_equal(t$name, "web_search")
t2 <- ns$.anthropic_web_search_tool(list(max_uses = 3, allowed_domains = "r-project.org",
                                         blocked_domains = "x.com",
                                         user_location = list(type = "approximate")))
expect_equal(t2$max_uses, 3L)
expect_equal(t2$allowed_domains[[1]], "r-project.org")
expect_equal(t2$blocked_domains[[1]], "x.com")
expect_equal(t2$user_location$type, "approximate")

# block parser: query from server_tool_use, citations from text blocks
content <- list(
    list(type = "server_tool_use", name = "web_search", input = list(query = "R version")),
    list(type = "web_search_tool_result", content = list()),
    list(type = "text", text = "R 4.6.0",
         citations = list(list(url = "https://www.r-project.org/", title = "R"))),
    list(type = "text", text = "more", citations = list()))
info <- ns$.anthropic_search_blocks(content)
expect_equal(length(info$searches), 1L)
expect_equal(info$searches[[1]]$query, "R version")
expect_equal(length(info$citations), 1L)
expect_equal(info$citations[[1]]$url, "https://www.r-project.org/")

# no search blocks -> empty
empty <- ns$.anthropic_search_blocks(list(list(type = "text", text = "hi")))
expect_equal(length(empty$citations), 0L)
expect_equal(length(empty$searches), 0L)

% tinyrox says don't edit this manually, but it can't stop you!
\name{prices_snapshot_stale}
\alias{prices_snapshot_stale}
\title{Is the bundled price snapshot stale?}
\usage{
prices_snapshot_stale(max_age_days = 90)
}
\arguments{
\item{max_age_days}{Numeric. Age threshold in days; default 90.}
}
\value{
`TRUE` when the bundled snapshot is older than
  `max_age_days`, otherwise `FALSE`.
}
\description{
Convenience wrapper over [prices_snapshot_date()] for staleness
alerts, so callers don't repeat the date arithmetic. Offline only;
it does not check the network for newer prices.
}
\examples{
prices_snapshot_stale()
prices_snapshot_stale(max_age_days = 30)
}
\seealso{
[prices_snapshot_date()]
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{llm_key}
\alias{llm_key}
\title{Set LLM API Key}
\usage{
llm_key(key)
}
\arguments{
\item{key}{Character. API key for authentication.}
}
\value{
The previous value of the option, invisibly.
}
\description{
Stores an API key in the \code{llm.api.api_key} option, which
\code{\link{chat}} prefers over environment variables.
}
\examples{
old <- llm_key("sk-not-a-real-key")
llm_key(old)  # restore
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mcp_start}
\alias{mcp_start}
\title{Start and connect to an MCP server}
\usage{
mcp_start(command, args = character(), port = NULL, name = NULL,
          startup_wait = 2)
}
\arguments{
\item{command}{Character. Command to run the server (e.g., "r", "Rscript").}

\item{args}{Character vector. Arguments (path to server script).}

\item{port}{Integer. Port for the server (default: random 7850-7899).}

\item{name}{Character. Friendly name.}

\item{startup_wait}{Numeric. Seconds to wait for server startup.}
}
\value{
An MCP connection object.
}
\description{
Spawns an MCP server process and connects to it.
Requires the server script to support --port argument.
}
\examples{
\dontrun{
conn <- mcp_start("Rscript", args = "mcp_server.R", port = 7850)
mcp_close(conn)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mcp_call}
\alias{mcp_call}
\title{Call a tool on an MCP server}
\usage{
mcp_call(conn, name, arguments = list())
}
\arguments{
\item{conn}{An MCP connection object.}

\item{name}{Character. Tool name.}

\item{arguments}{List. Tool arguments.}
}
\value{
Tool result (list with content and text).
}
\description{
Call a tool on an MCP server
}
\examples{
\dontrun{
conn <- mcp_connect(port = 7850)
result <- mcp_call(conn, "read_file", list(path = "README.md"))
mcp_close(conn)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{prices_snapshot_date}
\alias{prices_snapshot_date}
\title{Bundled price-snapshot date}
\usage{
prices_snapshot_date()
}
\value{
Character scalar in `YYYY-MM-DD` format.
}
\description{
`llm.api` estimates `usage$cost` from a model-price table baked into
the package at release time. This returns the date that bundled
table was generated.
}
\details{
It does not contact the network, check for newer prices, or update
the installed package. Use the date to display staleness warnings
(see [prices_snapshot_stale()]) or to decide when the package
maintainer should regenerate the snapshot for a future release.

The table is generated from BerriAI/litellm's
`model_prices_and_context_window.json`
(\url{https://github.com/BerriAI/litellm/blob/main/model_prices_and_context_window.json}).
Cost estimates are offline and approximate, and may differ from
current provider billing. Cached-input pricing follows each
provider's published model: OpenAI
(\url{https://platform.openai.com/docs/guides/prompt-caching}),
Moonshot (\url{https://www.kimi.com/help/kimi-api/api-pricing}), and
Anthropic
(\url{https://platform.claude.com/en/docs/build-with-claude/prompt-caching}).

}
\examples{
prices_snapshot_date()
}
\seealso{
[prices_snapshot_stale()], [usage_cost()]
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{chat_session}
\alias{chat_session}
\title{Create a stateful chat session}
\usage{
chat_session(model = NULL, system_prompt = NULL,
             provider = c("openai", "anthropic", "moonshot", "ollama"), ...)
}
\arguments{
\item{model}{Character. Model name.}

\item{system_prompt}{Character or NULL. System prompt.}

\item{provider}{Character. Provider: "openai", "anthropic", "moonshot",
or "ollama".}

\item{...}{Additional parameters passed to chat().}
}
\value{
A list with functions:
  \item{chat(message)}{Send a message, returns response text}
  \item{stream(message)}{Stream a response, returns response text}
  \item{stream_async(message)}{Async stream for shinychat (returns string)}
  \item{last_turn()}{Get the last response as list(role, text)}
  \item{history()}{Get full conversation history}
  \item{clear()}{Clear conversation history}
}
\description{
Creates a chat session that maintains conversation history internally.
Returns a list of functions for interacting with the session.
}
\examples{
\dontrun{
# Create a session
session <- chat_session(model = "gpt-5.4-mini", system_prompt = "You are helpful.")

# Chat (history maintained automatically)
response <- session$chat("Hello")
response2 <- session$chat("Tell me more")

# Get last response
last <- session$last_turn()
last$text

# Clear and start over
session$clear()
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{list_ollama_models}
\alias{list_ollama_models}
\title{List Ollama models}
\usage{
list_ollama_models(base_url = "http://localhost:11434")
}
\arguments{
\item{base_url}{Character. Ollama server URL (default: http://localhost:11434).}
}
\value{
A data frame with model information (name, size, modified).
}
\description{
List all models downloaded in Ollama.
}
\examples{
\dontrun{
list_ollama_models()
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{usage_cost}
\alias{usage_cost}
\title{Estimate the USD cost of one call's token usage}
\usage{
usage_cost(model, provider, usage)
}
\arguments{
\item{model}{Character. Model id as sent to the provider.}

\item{provider}{Character. "anthropic", "openai", "moonshot", or
"ollama".}

\item{usage}{A usage list as found in `chat()$usage` or
`agent()$usage`.}
}
\value{
Numeric scalar (USD), or `NA_real_` when `usage` is `NULL`,
  the model isn't in the snapshot, or cache reads can't be priced.
}
\description{
Computes the offline cost estimate for a usage object, the same
value `chat()` and `agent()` attach as `usage$cost`. Reads whichever
shape the provider returned (Anthropic's `input_tokens` /
`output_tokens`, or the OpenAI-compatible `prompt_tokens` /
`completion_tokens`) and accounts for prompt caching: Anthropic
cache writes/reads via published multipliers, OpenAI / Moonshot
cache hits at the bundled per-model `cache_read` rate. OpenAI cache
hits are read from `prompt_tokens_details$cached_tokens` on a raw
response, falling back to a flat `cache_read_input_tokens` field as
exposed by `agent()` aggregates, so an `agent()$usage` object
recomputes to the same cost it already carries.
}
\details{
Costs come from the bundled price snapshot, so they are offline,
approximate, and may differ from current provider billing. See
[prices_snapshot_date()].

}
\examples{
\dontrun{
r <- chat("hi", model = "claude-sonnet-4-6", cache = "5m")
usage_cost("claude-sonnet-4-6", "anthropic", r$usage)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{print.mcp_connection}
\alias{print.mcp_connection}
\title{Print an MCP Connection}
\usage{
\method{print}{mcp_connection}(x, ...)
}
\arguments{
\item{x}{An MCP connection object.}

\item{...}{Unused.}
}
\value{
\code{x}, invisibly. Called for the side effect of printing a
  summary of the connection state and available tools.
}
\description{
S3 print method for MCP connection objects.
}
\examples{
\dontrun{
conn <- mcp_connect(port = 7850)
print(conn)
mcp_close(conn)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mcp_connect}
\alias{mcp_connect}
\title{Connect to an MCP server}
\usage{
mcp_connect(host = "localhost", port, name = NULL, timeout = 30)
}
\arguments{
\item{host}{Character. Server hostname (default: "localhost").}

\item{port}{Integer. Server port.}

\item{name}{Character. Friendly name for this server.}

\item{timeout}{Numeric. Connection timeout in seconds (default: 30).}
}
\value{
An MCP connection object (list with socket and tools).
}
\description{
Connects to an MCP server via TCP socket.
}
\examples{
\dontrun{
# Start server first: r mcp_server.R --port 7850
conn <- mcp_connect(port = 7850, name = "codeR")
tools <- mcp_tools(conn)
result <- mcp_call(conn, "read_file", list(path = "README.md"))
mcp_close(conn)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{provider_default_model}
\alias{provider_default_model}
\title{Default model for a provider}
\usage{
provider_default_model(provider)
}
\arguments{
\item{provider}{Character. One of `"openai"`, `"anthropic"`,
`"moonshot"`, `"ollama"`.}
}
\value{
Character. The default model id for that provider.
}
\description{
Returns the model name `chat()` falls back to when the caller
doesn't specify one. Useful for client code that wants to display
the resolved model upfront (e.g., in a status line) without
duplicating the lookup table.
}
\examples{
provider_default_model("anthropic")
provider_default_model("moonshot")
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{chat}
\alias{chat}
\title{Chat with an LLM}
\usage{
chat(prompt, model = NULL, system = NULL, history = NULL, temperature = NULL,
     max_tokens = NULL,
     provider = c("auto", "openai", "anthropic", "moonshot", "ollama"),
     stream = FALSE, cache = c("none", "5m", "1h"),
     thinking_budget_tokens = NULL, ...)
}
\arguments{
\item{prompt}{Character. The user message to send.}

\item{model}{Character. Model name (e.g., "gpt-5.4-mini", "claude-sonnet-4-6", "qwen3.5:9b").}

\item{system}{Character or NULL. System prompt to set context.}

\item{history}{List or NULL. Previous conversation turns.}

\item{temperature}{Numeric or NULL. Sampling temperature (0-2).}

\item{max_tokens}{Integer or NULL. Maximum tokens in response.}

\item{provider}{Character. Provider: "auto", "openai", "anthropic",
"moonshot", or "ollama".}

\item{stream}{Logical. Stream the response (prints as it arrives).}

\item{cache}{Character. Anthropic prompt caching for the system
message: \code{"none"} (default), \code{"5m"}, or \code{"1h"}
ephemeral TTL. Anthropic-only; warns and degrades to \code{"none"}
for other providers.}

\item{thinking_budget_tokens}{Integer or NULL. Anthropic extended
thinking budget; must be at least 1024 and less than
\code{max_tokens}. Anthropic-only; ignored with a warning for
other providers.}

\item{...}{Additional parameters passed to the API.}
}
\value{
A list with:
  \item{content}{The assistant's response text}
  \item{thinking}{Chain-of-thought from reasoning models, or NULL.
    Populated from \code{reasoning_content} (DeepSeek, Moonshot Kimi,
    vLLM, SGLang), \code{reasoning} (OpenRouter), or Anthropic
    \code{thinking} blocks. Normalized across providers.}
  \item{finish_reason}{Why generation stopped. \code{"stop"} on a
    normal completion, \code{"length"} when truncated by max_tokens.
    A reasoning model that returns empty \code{content} with
    \code{finish_reason == "length"} ran out of budget mid-thought;
    raise \code{max_tokens}.}
  \item{model}{Model used}
  \item{usage}{Token usage (if available). When the model is in the
    bundled price snapshot, also carries \code{cost} as a USD scalar;
    Ollama is treated as free (\code{cost = 0}); unknown models leave
    \code{cost = NA_real_}. See \code{\link{prices_snapshot_date}}.}
  \item{history}{Updated conversation history}
}
\description{
Send a message to a Large Language Model and get a response.
}
\examples{
\dontrun{
# Simple chat
chat("What is 2+2?")

# With system prompt
chat("Explain R", system = "You are a helpful programming tutor.")

# Continue conversation
result <- chat("Hello")
chat("Tell me more", history = result$history)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{chat_claude}
\alias{chat_claude}
\title{Chat with Anthropic Claude}
\usage{
chat_claude(prompt, model = "claude-sonnet-4-6", ...)
}
\arguments{
\item{prompt}{Character. The user message to send.}

\item{model}{Character. Model name (e.g., "gpt-5.4-mini", "claude-sonnet-4-6", "qwen3.5:9b").}

\item{...}{Additional parameters passed to the API.}
}
\value{
The assistant's response as a character string, or a list when
  \code{history} is in use. See \code{\link{chat}} for details.
}
\description{
Convenience wrapper for 'Anthropic' 'Claude' models.
}
\examples{
\dontrun{
chat_claude("Explain the theory of relativity")
chat_claude("Write a poem", model = "claude-haiku-4-5")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{agent}
\alias{agent}
\title{Chat with tool use (agentic mode)}
\usage{
agent(prompt, tools = list(), tool_handler = NULL, system = NULL, model = NULL,
      provider = c("anthropic", "openai", "moonshot", "ollama"),
      max_turns = 20L, verbose = TRUE, history = NULL, history_callback = NULL,
      cache = c("none", "5m", "1h"), thinking_budget_tokens = NULL, ...)
}
\arguments{
\item{prompt}{Character. The user message.}

\item{tools}{List. Tool definitions (from mcp_tools_for_claude or manual).}

\item{tool_handler}{Function. Called with (name, args), returns result string.}

\item{system}{Character. System prompt.}

\item{model}{Character. Model name.}

\item{provider}{Character. Provider: "anthropic", "openai", "moonshot",
or "ollama".}

\item{max_turns}{Integer. Maximum tool-use turns (default: 20).}

\item{verbose}{Logical. Print tool calls and results.}

\item{history}{List or NULL. Previous conversation history to continue from.}

\item{history_callback}{Function or NULL. Called as
\code{history_callback(history)} after each assistant message is
appended and after each tool result is appended. Lets callers
snapshot intermediate state so an interrupt mid-turn doesn't lose
the work that was already done. Errors raised inside the callback
are swallowed so telemetry/snapshotting can't break a turn.}

\item{cache}{Character. Anthropic prompt caching for the system
message: \code{"none"} (default), \code{"5m"}, or \code{"1h"}
ephemeral TTL. Anthropic-only; warns and degrades to \code{"none"}
for other providers.}

\item{thinking_budget_tokens}{Integer or NULL. Anthropic extended
thinking budget; must be at least 1024 and less than
\code{max_tokens}. Anthropic-only; ignored with a warning for
other providers.}

\item{...}{Additional parameters passed to the API.}
}
\value{
List with final response and conversation history. The
  returned \code{$usage} carries cumulative \code{input_tokens},
  \code{output_tokens}, \code{total_tokens}, and \code{cost} (USD
  scalar, derived from the bundled price snapshot; \code{0} for
  Ollama; \code{NA_real_} for models not in the snapshot). It also
  carries cumulative cache activity: \code{cache_read_input_tokens}
  (Anthropic cache reads plus OpenAI/Moonshot cached prompt tokens),
  \code{cache_creation_input_tokens} (total Anthropic cache writes),
  and the per-TTL split \code{cache_creation$ephemeral_5m_input_tokens}
  / \code{cache_creation$ephemeral_1h_input_tokens}. Passing this
  \code{$usage} back to \code{\link{usage_cost}} recomputes the same
  \code{cost}.
}
\description{
Send a prompt to an LLM with tools. Automatically handles tool calls
in a loop until the model responds with text only.
}
\examples{
\dontrun{
# With MCP server
conn <- mcp_connect("r", "mcp_server.R")
tools <- mcp_tools_for_claude(conn)

result <- agent(
  "What files are in the current directory?",
  tools = tools,
  tool_handler = function(name, args) {
    mcp_call(conn, name, args)$text
  }
)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mcp_close}
\alias{mcp_close}
\title{Close an MCP connection}
\usage{
mcp_close(conn)
}
\arguments{
\item{conn}{An MCP connection object.}
}
\value{
\code{NULL}, invisibly. Called for its side effect of closing
  the underlying socket.
}
\description{
Close an MCP connection
}
\examples{
\dontrun{
conn <- mcp_connect(port = 7850)
mcp_close(conn)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{chat_session_anthropic}
\alias{chat_session_anthropic}
\title{Create Anthropic chat session}
\usage{
chat_session_anthropic(model = "claude-sonnet-4-6", system_prompt = NULL, ...)
}
\arguments{
\item{model}{Character. Model name (default: "claude-sonnet-4-6").}

\item{system_prompt}{Character or NULL. System prompt.}

\item{...}{Additional parameters passed to chat_session().}
}
\value{
A chat session list.
}
\description{
Convenience wrapper for chat_session with Anthropic provider.
}
\examples{
# Construct a session (no network call yet)
session <- chat_session_anthropic(system_prompt = "You are helpful.")
\dontrun{
session$chat("Hello")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{chat_session_openai}
\alias{chat_session_openai}
\title{Create OpenAI chat session}
\usage{
chat_session_openai(model = "gpt-5.4-mini", system_prompt = NULL, ...)
}
\arguments{
\item{model}{Character. Model name (default: "gpt-5.4-mini").}

\item{system_prompt}{Character or NULL. System prompt.}

\item{...}{Additional parameters passed to chat_session().}
}
\value{
A chat session list.
}
\description{
Convenience wrapper for chat_session with OpenAI provider.
}
\examples{
# Construct a session (no network call yet)
session <- chat_session_openai(system_prompt = "You are helpful.")
\dontrun{
session$chat("Hello")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{create_agent}
\alias{create_agent}
\title{Create an agent with MCP servers}
\usage{
create_agent(servers = list(), system = NULL, model = NULL,
             provider = c("anthropic", "openai", "moonshot", "ollama"),
             verbose = TRUE)
}
\arguments{
\item{servers}{Named list of server configs. Each can be:
- `list(port = 7850)` for already-running servers
- `list(command = "r", args = "server.R", port = 7850)` to start and connect}

\item{system}{Character. Default system prompt.}

\item{model}{Character. Default model.}

\item{provider}{Character. Provider: "anthropic", "openai", "moonshot",
or "ollama".}

\item{verbose}{Logical. Print tool calls.}
}
\value{
A function that takes a prompt and returns a response.
}
\description{
Convenience function that sets up MCP connections and returns
a function for chatting with tools.
}
\examples{
\dontrun{
# Connect to already-running server
chat_fn <- create_agent(
  servers = list(codeR = list(port = 7850)),
  system = "You are a helpful coding assistant."
)

# Or start server automatically
chat_fn <- create_agent(
  servers = list(
    codeR = list(command = "r", args = "mcp_server.R", port = 7850)
  )
)

result <- chat_fn("List files in current directory")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{llm_base}
\alias{llm_base}
\title{Set LLM API Base URL}
\usage{
llm_base(url)
}
\arguments{
\item{url}{Character. Base URL for the API endpoint.}
}
\value{
The previous value of the option, invisibly.
}
\description{
Stores a base URL in the \code{llm.api.api_base} option, which
\code{\link{chat}} uses as the default endpoint.
}
\examples{
old <- llm_base("http://localhost:11434")  # 'Ollama'
llm_base(old)  # restore
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mcp_tools_for_claude}
\alias{mcp_tools_for_claude}
\title{Format MCP tools for Claude API}
\usage{
mcp_tools_for_claude(conn)
}
\arguments{
\item{conn}{An MCP connection, or list of connections.}
}
\value{
List of tools in API format.
}
\description{
Wrapper for \code{\link{mcp_tools_for_api}}, retained for backwards
compatibility.
}
\examples{
\dontrun{
conn <- mcp_connect(host = "localhost", port = 7850)
tools <- mcp_tools_for_claude(conn)
mcp_close(conn)
}
}
